#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the Audyssey settings of Denon AVR receivers.

:copyright: (c) 2020 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import logging

from typing import Hashable, List, Optional

import attr

from .appcommand import AppCommandCmd, AppCommandCmdParam, AppCommands
from .const import (
    DENON_ATTR_SETATTR, DYNAMIC_VOLUME_MAP, DYNAMIC_VOLUME_MAP_LABELS,
    MULTI_EQ_MAP, MULTI_EQ_MAP_LABELS, REF_LVL_OFFSET_MAP,
    REF_LVL_OFFSET_MAP_LABELS)
from .foundation import DenonAVRFoundation, convert_string_int_bool
from .exceptions import AvrCommandError, AvrProcessingError

_LOGGER = logging.getLogger(__name__)


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRAudyssey(DenonAVRFoundation):
    """Audyssey Settings."""

    _dynamiceq: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool),
        default=None)
    _dynamiceq_control: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool),
        default=None)
    _reflevoffset: Optional[str] = attr.ib(
        converter=attr.converters.optional(REF_LVL_OFFSET_MAP.get),
        default=None)
    _reflevoffset_control: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool),
        default=None)
    _dynamicvol: Optional[str] = attr.ib(
        converter=attr.converters.optional(DYNAMIC_VOLUME_MAP.get),
        default=None)
    _dynamicvol_control: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool),
        default=None)
    _multeq: Optional[str] = attr.ib(
        converter=attr.converters.optional(MULTI_EQ_MAP.get),
        default=None)
    _multeq_control: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool),
        default=None)

    # Update tags for attributes
    # AppCommand0300.xml interface
    appcommand0300_attrs = {
        AppCommands.GetAudyssey: None}

    def setup(self) -> None:
        """Ensure that the instance is initialized."""
        # Add tags for a potential AppCommand.xml update
        for tag in self.appcommand0300_attrs:
            self._device.api.add_appcommand0300_update_tag(tag)

        self._is_setup = True

    async def async_update(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None) -> None:
        """Update Audyssey asynchronously."""
        # Ensure instance is setup before updating
        if self._is_setup is False:
            self.setup()

        # Update state
        await self.async_update_audyssey(
            global_update=global_update, cache_id=cache_id)

    async def async_update_audyssey(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None):
        """Update Audyssey status of device."""
        if self._device.use_avr_2016_update is True:
            await self.async_update_attrs_appcommand(
                self.appcommand0300_attrs,
                appcommand0300=True,
                global_update=global_update,
                cache_id=cache_id,
                ignore_missing_response=True)
        elif self._device.use_avr_2016_update is False:
            # Not available
            pass
        else:
            raise AvrProcessingError(
                "Device is not setup correctly, update method not set")

    async def _async_set_audyssey(self, cmd: AppCommandCmd) -> None:
        """Set Audyssey parameter."""
        res = await self._device.api.async_post_appcommand(
            self._device.urls.appcommand0300, (cmd,))

        try:
            if res.find("cmd").text != "OK":
                raise AvrProcessingError(
                    "SetAudyssey command {} failed".format(cmd.name))
        except AttributeError as err:
            raise AvrProcessingError(
                "SetAudyssey command {} failed".format(cmd.name)) from err

    ##############
    # Properties #
    ##############
    @property
    def dynamic_eq(self) -> Optional[bool]:
        """Return value of Dynamic EQ."""
        return self._dynamiceq

    @property
    def reference_level_offset(self) -> Optional[str]:
        """Return value of Reference Level Offset."""
        return self._reflevoffset

    @property
    def reference_level_offset_setting_list(self) -> List[str]:
        """Return a list of available reference level offset settings."""
        return list(REF_LVL_OFFSET_MAP_LABELS.keys())

    @property
    def dynamic_volume(self) -> Optional[str]:
        """Return value of Dynamic Volume."""
        return self._dynamicvol

    @property
    def dynamic_volume_setting_list(self) -> List[str]:
        """Return a list of available Dynamic Volume settings."""
        return list(DYNAMIC_VOLUME_MAP_LABELS.keys())

    @property
    def multi_eq(self) -> Optional[str]:
        """Return value of MultiEQ."""
        return self._multeq

    @property
    def multi_eq_setting_list(self) -> List[str]:
        """Return a list of available MultiEQ settings."""
        return list(MULTI_EQ_MAP_LABELS.keys())

    ##########
    # Setter #
    ##########
    async def async_dynamiceq_off(self) -> None:
        """Turn DynamicEQ off."""
        cmd = attr.evolve(
            AppCommands.SetAudysseyDynamicEQ, param_list=(
                AppCommandCmdParam(name="dynamiceq", text=0),))
        await self._async_set_audyssey(cmd)

    async def async_dynamiceq_on(self) -> None:
        """Turn DynamicEQ on."""
        cmd = attr.evolve(
            AppCommands.SetAudysseyDynamicEQ, param_list=(
                AppCommandCmdParam(name="dynamiceq", text=1),))
        await self._async_set_audyssey(cmd)

    async def async_set_multieq(self, value: str) -> None:
        """Set MultiEQ mode."""
        setting = MULTI_EQ_MAP_LABELS.get(value)
        if setting is None:
            raise AvrCommandError(
                "Value {} not known for MultiEQ".format(value))
        cmd = attr.evolve(
            AppCommands.SetAudysseyMultiEQ, param_list=(
                AppCommandCmdParam(name="multeq", text=setting),))
        await self._async_set_audyssey(cmd)

    async def async_set_reflevoffset(self, value: str) -> None:
        """Set Reference Level Offset."""
        # Reference level offset can only be used with DynamicEQ
        if self._dynamiceq is False:
            raise AvrCommandError(
                "Reference level could only be set when DynamicEQ is active")
        setting = REF_LVL_OFFSET_MAP_LABELS.get(value)
        if setting is None:
            raise AvrCommandError(
                "Value {} not known for Reference level offset".format(value))
        cmd = attr.evolve(
            AppCommands.SetAudysseyReflevoffset, param_list=(
                AppCommandCmdParam(name="reflevoffset", text=setting),))
        await self._async_set_audyssey(cmd)

    async def async_set_dynamicvol(self, value: str) -> None:
        """Set Dynamic Volume."""
        setting = DYNAMIC_VOLUME_MAP_LABELS.get(value)
        if setting is None:
            raise AvrCommandError(
                "Value {} not known for Dynamic Volume".format(value))
        cmd = attr.evolve(
            AppCommands.SetAudysseyDynamicvol, param_list=(
                AppCommandCmdParam(name="dynamicvol", text=setting),))
        await self._async_set_audyssey(cmd)

    async def async_toggle_dynamic_eq(self) -> None:
        """Toggle DynamicEQ."""
        if self._dynamiceq is True:
            await self.async_dynamiceq_off()
        else:
            await self.async_dynamiceq_on()


def audyssey_factory(instance: DenonAVRFoundation) -> DenonAVRAudyssey:
    """Create  DenonAVRAudyssey at receiver instances."""
    # pylint: disable=protected-access
    new = DenonAVRAudyssey(device=instance._device)
    return new



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the foundation classes for Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import logging
import xml.etree.ElementTree as ET

from copy import deepcopy
from typing import Dict, Hashable, List, Optional

import attr
import httpx

from .appcommand import AppCommandCmd, AppCommands
from .api import DenonAVRApi
from .exceptions import (
    AvrForbiddenError, AvrProcessingError, AvrRequestError, AvrTimoutError,
    AvrInvalidResponseError)
from .const import (
    APPCOMMAND_CMD_TEXT, APPCOMMAND_NAME, AVR, AVR_X, AVR_X_2016,
    DENON_ATTR_SETATTR, DENONAVR_URLS, DESCRIPTION_TYPES,
    DEVICEINFO_AVR_X_PATTERN, DEVICEINFO_COMMAPI_PATTERN, MAIN_ZONE,
    VALID_RECEIVER_TYPES, VALID_ZONES, ReceiverType, ReceiverURLs, ZONE2,
    ZONE2_URLS, ZONE3, ZONE3_URLS)
from .ssdp import evaluate_scpd_xml


_LOGGER = logging.getLogger(__name__)


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRDeviceInfo:
    """Implements a class with device information of the receiver."""

    api: DenonAVRApi = attr.ib(
        validator=attr.validators.instance_of(DenonAVRApi),
        default=attr.Factory(DenonAVRApi),
        kw_only=True)
    receiver: Optional[ReceiverType] = attr.ib(
        validator=attr.validators.optional(
            attr.validators.in_(VALID_RECEIVER_TYPES)),
        default=None)
    urls: ReceiverURLs = attr.ib(
        validator=attr.validators.instance_of(ReceiverURLs), init=False)
    zone: str = attr.ib(
        validator=attr.validators.in_(VALID_ZONES),
        default=MAIN_ZONE,
        kw_only=True)
    friendly_name: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None)
    manufacturer: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None)
    model_name: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None)
    serial_number: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None)
    use_avr_2016_update: Optional[bool] = attr.ib(
        converter=attr.converters.optional(bool), default=None)
    _power: Optional[str] = attr.ib(
        converter=attr.converters.optional(str),
        default=None)
    _is_setup: bool = attr.ib(converter=bool, default=False, init=False)
    _allow_recovery: bool = attr.ib(
        converter=bool, default=False, init=False)

    def __attrs_post_init__(self) -> None:
        """Initialize special attributes."""
        # URLs depending from value of self.zone attribute
        if self.zone == MAIN_ZONE:
            self.urls = DENONAVR_URLS
        elif self.zone == ZONE2:
            self.urls = ZONE2_URLS
        elif self.zone == ZONE3:
            self.urls = ZONE3_URLS
        else:
            raise ValueError("Invalid zone {}".format(self.zone))

    def get_own_zone(self):
        """
        Get zone from actual instance.

        These zone information are used to evaluate responses of HTTP POST
        commands.
        """
        if self.zone == MAIN_ZONE:
            return "zone1"
        return self.zone.lower()

    async def async_setup(self) -> None:
        """Ensure that configuration is loaded from receiver asynchronously."""
        # Own setup
        await self.async_identify_receiver()
        await self.async_identify_update_method()
        await self.async_get_device_info()

        # Add tags for a potential AppCommand.xml update
        self.api.add_appcommand_update_tag(AppCommands.GetAllZonePowerStatus)

        self._is_setup = True

    async def async_update(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None) -> None:
        """Update status asynchronously."""
        # Ensure instance is setup before updating
        if self._is_setup is False:
            await self.async_setup()

        # Update power status
        await self.async_update_power(
            global_update=global_update, cache_id=cache_id)

    async def async_identify_receiver(self) -> None:
        """Identify receiver asynchronously."""
        # Test Deviceinfo.xml if receiver is a AVR-X with port 80 for pre 2016
        # devices and port 8080 devices 2016 and later
        # 2016 models has also some of the XML but not all, try first 2016
        r_types = [AVR_X, AVR_X_2016]

        for r_type in r_types:
            self.api.port = r_type.port
            # This XML is needed to get the sources of the receiver
            try:
                # Deviceinfo.xml is static and can be cached for the whole time
                xml = await self.api.async_get_xml(
                    self.urls.deviceinfo, cache_id=0)
            except AvrTimoutError as err:
                _LOGGER.debug(
                    "Timeout when identifying receiver", exc_info=err)

                # Raise error only when ran over all models
                if r_types.index(r_type) == len(r_types) - 1:
                    raise

            except AvrRequestError as err:
                _LOGGER.debug(
                    "Request error when identifying receiver", exc_info=err)
            else:
                is_avr_x = self._is_avr_x(xml)
                if is_avr_x:
                    self.receiver = r_type
                    # Receiver identified, return
                    return

        # If check of Deviceinfo.xml was not successfull, receiver is type AVR
        self.receiver = AVR
        self.api.port = AVR.port

    @staticmethod
    def _is_avr_x(deviceinfo: ET.Element) -> bool:
        """Evaluate Deviceinfo.xml if the device is a AVR-X device."""
        # First test by CommApiVers
        try:
            if bool(DEVICEINFO_COMMAPI_PATTERN.search(
                    deviceinfo.find("CommApiVers").text) is not None):
                # receiver found , return True
                return True
        except AttributeError:
            # AttributeError occurs when ModelName tag is not found.
            # In this case there is no AVR-X device
            pass

        # if first test did not find AVR-X device, check by model name
        try:
            if bool(DEVICEINFO_AVR_X_PATTERN.search(
                    deviceinfo.find("ModelName").text) is not None):
                # receiver found , return True
                return True
        except AttributeError:
            # AttributeError occurs when ModelName tag is not found
            # In this case there is no AVR-X device
            pass

        return False

    async def async_identify_update_method(self) -> None:
        """
        Identify the correct update method for the receiver asynchronously.

        As a result this method sets friendly name too.
        """
        # AVR receivers do not support AppCommand.xml interface
        if self.receiver == AVR:
            self.use_avr_2016_update = False
            _LOGGER.info("AVR device, AppCommand.xml interface not supported")
        else:
            try:
                xml = await self.api.async_post_appcommand(
                    self.urls.appcommand,
                    (AppCommands.GetFriendlyName,))
            except AvrTimoutError as err:
                _LOGGER.debug(
                    "Timeout when identifying update method", exc_info=err)
                raise
            except AvrRequestError as err:
                _LOGGER.debug(
                    "Request error when identifying update method",
                    exc_info=err)
                self.use_avr_2016_update = False
                _LOGGER.info(
                    "AVR-X device, AppCommand.xml interface not supported")
            else:
                self.use_avr_2016_update = True
                _LOGGER.info("AVR-X device, using AppCommand.xml interface")
                self._set_friendly_name(xml)

            if self.use_avr_2016_update is False:
                try:
                    xml = await self.api.async_get_xml(self.urls.mainzone)
                except AvrTimoutError as err:
                    _LOGGER.debug(
                        "Timeout when identifying update method", exc_info=err)
                    raise
                except AvrRequestError as err:
                    _LOGGER.debug(
                        "Request error getting friendly name", exc_info=err)
                    _LOGGER.info(
                        "Receiver name could not be determined. Using standard"
                        " name: Denon AVR")
                    self.friendly_name = "Denon AVR"
                else:
                    self._set_friendly_name(xml)

    async def async_verify_avr_2016_update_method(
            self, cache_id: Hashable = None) -> None:
        """Verify if avr 2016 update method is working with all tags."""
        # AVR receivers do not support AppCommand.xml interface
        if self.receiver == AVR:
            self.use_avr_2016_update = False
        else:
            try:
                await self.api.async_get_global_appcommand(cache_id=cache_id)
            except AvrTimoutError as err:
                _LOGGER.debug(
                    "Timeout when verifying update method", exc_info=err)
                raise
            except AvrForbiddenError:
                # Recovery in case receiver changes port from 80 to 8080 which
                # might happen at Denon AVR-X 2016 receivers
                if self._allow_recovery is True:
                    self._allow_recovery = False
                    _LOGGER.warning(
                        "AppCommand.xml returns HTTP status 403. Running setup"
                        " again once to check if receiver interface switched "
                        "ports")
                    self._is_setup = False
                    await self.async_setup()
                    await self.async_verify_avr_2016_update_method(
                        cache_id=cache_id)
                else:
                    raise
            except AvrRequestError as err:
                _LOGGER.debug(
                    "Request error when verifying update method", exc_info=err)
                if self.use_avr_2016_update is True:
                    _LOGGER.error(
                        "Error verifying update method, deactivate "
                        "AppCommand.xml interface")
                self.use_avr_2016_update = False
            else:
                self._allow_recovery = True
                self.use_avr_2016_update = True

    def _set_friendly_name(self, xml: ET.Element) -> None:
        """Set FriendlyName from result xml."""
        # friendlyname tag of AppCommand.xml, FriendlyName tag main zone xml
        tags = ("./cmd/friendlyname", "./FriendlyName/value")
        for tag in tags:
            try:
                name = xml.find(tag).text
            except AttributeError:
                pass
            else:
                if name is not None:
                    self.friendly_name = name.strip()
                    break
        if self.friendly_name is None:
            _LOGGER.error(
                "No FriendlyName found, using standard name: Denon AVR")
            self.friendly_name = "Denon AVR"

    async def async_get_device_info(self) -> None:
        """Get device information."""
        port = DESCRIPTION_TYPES[self.receiver.type].port
        command = DESCRIPTION_TYPES[self.receiver.type].url
        url = "http://{host}:{port}{command}".format(
            host=self.api.host, port=port, command=command)

        device_info = None
        try:
            res = await self.api.async_get(command, port=port)
        except httpx.TimeoutException as err:
            _LOGGER.debug("Timeout when identifying receiver", exc_info=err)
            raise AvrTimoutError(
                "Timeout when identifying receiver", command) from err
        except httpx.HTTPError as err:
            _LOGGER.error(
                "During DenonAVR device identification, when trying to request"
                " %s the following error occurred: %s", url, err)
        else:
            device_info = evaluate_scpd_xml(url, res.text)

        if device_info is None:
            self.manufacturer = "Denon"
            self.model_name = "Unknown"
            self.serial_number = None
            _LOGGER.debug("No device info")
            raise AvrInvalidResponseError(
                "Unable to get device information of host {}, Device is in a "
                "corrupted state. Disconnect and reconnect power to the device"
                " and try again.".format(self.api.host), command)

        self.manufacturer = device_info["manufacturer"]
        self.model_name = device_info["modelName"]
        self.serial_number = device_info["serialNumber"]

    async def async_update_power(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None):
        """Update power status of device."""
        if self.use_avr_2016_update is True:
            await self.async_update_power_appcommand(
                global_update=global_update, cache_id=cache_id)
        elif self.use_avr_2016_update is False:
            await self.async_update_power_status_xml(cache_id=cache_id)
        else:
            raise AvrProcessingError(
                "Device is not setup correctly, update method not set")

    async def async_update_power_appcommand(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None):
        """Update power status from AppCommand.xml."""
        # Collect tags for AppCommand.xml call
        update_attrs = {"_power": AppCommands.GetAllZonePowerStatus}
        tags = tuple(i for i in update_attrs.values())
        # Execute call
        try:
            if global_update:
                xml = await self.api.async_get_global_appcommand(
                    cache_id=cache_id)
            else:
                xml = await self.api.async_post_appcommand(
                    self.urls.appcommand, tags, cache_id=cache_id)
        except AvrRequestError as err:
            _LOGGER.debug(
                "Error when getting power status", exc_info=err)
            raise

        # Extract relevant information
        zone = self.get_own_zone()

        attrs = deepcopy(update_attrs)
        for name, tag in attrs.items():
            try:
                # Check if attribute exists
                getattr(self, name)
                # Set new value
                setattr(
                    self,
                    name,
                    xml.find("./cmd[@{attribute}='{cmd}']/{zone}".format(
                        attribute=APPCOMMAND_CMD_TEXT,
                        cmd=tag.cmd_text,
                        zone=zone)).text)
                # Done
                update_attrs.pop(name, None)

            except (AttributeError, IndexError) as err:
                _LOGGER.debug(
                    "Failed updating attribute %s for zone %s", name,
                    self.zone, exc_info=err)

        # Check if each attribute was updated
        if update_attrs:
            raise AvrProcessingError(
                "Not all attributes updated, those are left: {}".format(
                    update_attrs))

    async def async_update_power_status_xml(
            self,
            cache_id: Optional[Hashable] = None):
        """Update power status from status xml."""
        # URLs to be scanned
        urls = [self.urls.status, self.urls.mainzone]
        # Variables with their tags to be updated
        update_attrs = {"_power": "./Power/value"}

        for url in urls:
            try:
                xml = await self.api.async_get_xml(
                    url, cache_id=cache_id)
            except AvrRequestError as err:
                _LOGGER.debug(
                    "Error when getting power status from url %s", url,
                    exc_info=err)
                continue
            attrs = deepcopy(update_attrs)
            for name, tag in attrs.items():
                try:
                    # Check if attribute exists
                    getattr(self, name)
                    # Set new value
                    setattr(self, name, xml.find(tag).text)
                    # Done
                    update_attrs.pop(name, None)
                except (AttributeError, IndexError) as err:
                    _LOGGER.debug(
                        "Failed updating attribute %s for zone %s", name,
                        self.zone, exc_info=err)

            # All done, no need for continuing
            if not update_attrs:
                break

        # Check if each attribute was updated
        if update_attrs:
            raise AvrProcessingError(
                "Not all attributes updated, those are left: {}".format(
                    update_attrs))

    ##############
    # Properties #
    ##############
    @property
    def power(self) -> Optional[str]:
        """
        Return the power state of the device.

        Possible values are: "ON", "STANDBY" and "OFF"
        """
        return self._power

    ##########
    # Setter #
    ##########

    async def async_power_on(self) -> None:
        """Turn on receiver via HTTP get command."""
        await self.api.async_get_command(self.urls.command_power_on)

    async def async_power_off(self) -> None:
        """Turn off receiver via HTTP get command."""
        await self.api.async_get_command(self.urls.command_power_standby)


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRFoundation:
    """
    Implements the foundation class of DenonAVR functions.

    All functions like the receiver, zones, sound modes, media settings, tone
    control use this class.
    """

    _device: DenonAVRDeviceInfo = attr.ib(
        validator=attr.validators.instance_of(DenonAVRDeviceInfo),
        default=attr.Factory(DenonAVRDeviceInfo),
        kw_only=True)
    _is_setup: bool = attr.ib(converter=bool, default=False, init=False)

    async def async_update_attrs_appcommand(
            self,
            update_attrs: Dict[AppCommandCmd, None],
            appcommand0300: bool = False,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None,
            ignore_missing_response: bool = False):
        """Update attributes from AppCommand.xml."""
        # Copy that we do not accidently change the wrong dict
        update_attrs = deepcopy(update_attrs)
        # Collect tags for AppCommand.xml call
        tags = tuple(i for i in update_attrs.keys())
        # Execute call
        try:
            if global_update:
                xml = await self._device.api.async_get_global_appcommand(
                    appcommand0300=appcommand0300, cache_id=cache_id)
            else:
                # Determine endpoint
                if appcommand0300:
                    url = self._device.urls.appcommand0300
                else:
                    url = self._device.urls.appcommand
                xml = await self._device.api.async_post_appcommand(
                    url, tags, cache_id=cache_id)
        except AvrRequestError as err:
            _LOGGER.debug(
                "Error when getting status update", exc_info=err)
            raise

        # Extract relevant information
        zone = self._device.get_own_zone()

        attrs = deepcopy(update_attrs)
        for app_command in attrs.keys():
            search_strings = self.create_appcommand_search_strings(
                app_command, zone)
            start = 0
            success = 0
            for i, pattern in enumerate(app_command.response_pattern):
                try:
                    start += 1
                    # Check if attribute exists
                    getattr(self, pattern.update_attribute)
                    # Set new value either from XML attribute or text
                    if pattern.get_xml_attribute is not None:
                        set_value = xml.find(
                            search_strings[i]).get(pattern.get_xml_attribute)
                    else:
                        set_value = xml.find(search_strings[i]).text

                    setattr(
                        self,
                        pattern.update_attribute,
                        set_value)
                    success += 1

                    _LOGGER.debug(
                        "Changing variable %s to value %s",
                        pattern.update_attribute, set_value)

                except (AttributeError, IndexError) as err:
                    _LOGGER.debug(
                        "Failed updating attribute %s for zone %s",
                        pattern.update_attribute, self._device.zone,
                        exc_info=err)

            if start == success:
                # Done
                update_attrs.pop(app_command, None)

        # Check if each attribute was updated
        if update_attrs and ignore_missing_response is False:
            raise AvrProcessingError(
                "Not all attributes updated, those are left: {}".format(
                    update_attrs))

    async def async_update_attrs_status_xml(
            self,
            update_attrs: Dict[str, str],
            urls: List[str],
            cache_id: Optional[Hashable] = None,
            ignore_missing_response: bool = False):
        """
        Update attributes from status xml.

        # URLs to be scanned. Like:
        urls = [self._device.urls.status, self._device.urls.mainzone]

        # Variables with their tags to be updated.
        # Key = Variable, Value = XML tag Like:
        update_attrs = {"power": "./Power/value"}
        """
        # Copy that we do not accidently change the wrong dict
        update_attrs = deepcopy(update_attrs)

        for url in urls:
            try:
                xml = await self._device.api.async_get_xml(
                    url, cache_id=cache_id)
            except AvrRequestError as err:
                _LOGGER.debug(
                    "Error when getting status update from url %s", url,
                    exc_info=err)
                continue
            attrs = deepcopy(update_attrs)
            for name, tag in attrs.items():
                try:
                    # Check if attribute exists
                    getattr(self, name)
                    # Set new value
                    setattr(self, name, xml.find(tag).text)
                    # Done
                    update_attrs.pop(name, None)

                    _LOGGER.debug(
                        "Changing variable %s to value %s", name,
                        xml.find(tag).text)

                except (AttributeError, IndexError) as err:
                    _LOGGER.debug(
                        "Failed updating attribute %s for zone %s", name,
                        self._device.zone, exc_info=err)

            # All done, no need for continuing
            if not update_attrs:
                break

        # Check if each attribute was updated
        if update_attrs and ignore_missing_response is False:
            raise AvrProcessingError(
                "Not all attributes updated, those are left: {}".format(
                    update_attrs))

    @staticmethod
    def create_appcommand_search_strings(
            app_command_cmd: AppCommandCmd, zone: str) -> List[str]:
        """Create search pattern for AppCommand(0300).xml response."""
        result = []

        for resp in app_command_cmd.response_pattern:
            string = "./cmd"
            # Text of cmd tag in query was added as attribute to response
            if app_command_cmd.cmd_text:
                string = string + "[@{}='{}']".format(
                    APPCOMMAND_CMD_TEXT, app_command_cmd.cmd_text)
            # Text of name tag in query was added as attribute to response
            if app_command_cmd.name:
                string = string + "[@{}='{}']".format(
                    APPCOMMAND_NAME, app_command_cmd.name)
            # Some results include a zone tag
            if resp.add_zone:
                string = string + "/{}".format(zone)
            # Suffix like /status, /volume
            string = string + resp.suffix

            # A complete search string with all strributes set looks like
            # ./cmd[@cmd_text={cmd_text}][@name={name}]/zone1/volume
            result.append(string)

        return result


def set_api_host(
        instance: DenonAVRFoundation,
        attribute: attr.Attribute,
        value: str) -> str:
    """Change API host on host changes too."""
    # First change _device.api.host then return value
    instance._device.api.host = value  # pylint: disable=protected-access
    return value


def set_api_timeout(
        instance: DenonAVRFoundation,
        attribute: attr.Attribute,
        value: float) -> float:
    """Change API timeout on timeout changes too."""
    # First change _device.api.host then return value
    timeout = httpx.Timeout(value, read=max(value, 15.0))
    instance._device.api.timeout = timeout  # pylint: disable=protected-access
    return value


def convert_string_int_bool(value: str) -> bool:
    """Convert an integer from string format to bool."""
    if value is None:
        return None
    return bool(int(value))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the REST API to Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""


class DenonAvrError(Exception):
    """Define an error for Denon AVR."""


class AvrCommandError(DenonAvrError):
    """Define an error command errors."""

    # pylint: disable=useless-super-delegation
    def __init__(self, message: str, *args, **kwargs) -> None:
        """Create a new instance."""
        super().__init__(message, *args, **kwargs)


class AvrProcessingError(DenonAvrError):
    """Define an error for process errors."""

    # pylint: disable=useless-super-delegation
    def __init__(self, message: str, *args, **kwargs) -> None:
        """Create a new instance."""
        super().__init__(message, *args, **kwargs)


class AvrRequestError(DenonAvrError):
    """Define an error related to a HTTP request for Denon AVR."""

    def __init__(self, message: str, request: str, *args, **kwargs) -> None:
        """Create a new instance."""
        self.request = request
        super().__init__(message, *args, **kwargs)


class AvrTimoutError(AvrRequestError):
    """Define an error for timeouts during a HTTP request for Denon AVR."""


class AvrInvalidResponseError(AvrRequestError):
    """Define an error for invalid responses of Denon AVR."""


class AvrForbiddenError(AvrRequestError):
    """Define an error for forbidden endpoints (HTTP 403) of Denon AVR."""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the handler for sound mode of Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

from copy import deepcopy
import logging

from collections import OrderedDict
from typing import Dict, Hashable, Optional

import attr

from .appcommand import AppCommands
from .const import (
    AVR_X, AVR_X_2016, ALL_ZONE_STEREO, DENON_ATTR_SETATTR, SOUND_MODE_MAPPING)
from .exceptions import AvrProcessingError
from .foundation import DenonAVRFoundation


_LOGGER = logging.getLogger(__name__)


def rstrip_string(value: Optional[str]) -> Optional[str]:
    """Perform HTML unescape on value."""
    if value is None:
        return value
    return (str(value)).rstrip()


def sound_mode_rev_map_factory(instance: DenonAVRFoundation) -> Dict[str, str]:
    """
    Construct the sound_mode_rev_map.

    Reverse the key value structure. The sound_mode_rev_map is bigger,
    but allows for direct matching using a dictionary key access.
    The sound_mode_map is uses externally to set this dictionary
    because that has a nicer syntax.
    """
    mode_map = list(
        instance._sound_mode_map.items())  # pylint: disable=protected-access
    mode_map_rev = {}
    for matched_mode, sublist in mode_map:
        for raw_mode in sublist:
            mode_map_rev[raw_mode.upper()] = matched_mode
    return mode_map_rev


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRSoundMode(DenonAVRFoundation):
    """This class implements sound mode functions of Denon AVR receiver."""

    _support_sound_mode: Optional[bool] = attr.ib(
        converter=attr.converters.optional(bool),
        default=None)
    _sound_mode_raw: Optional[str] = attr.ib(
        converter=attr.converters.optional(rstrip_string),
        default=None)
    _sound_mode_map: Dict[str, str] = attr.ib(  # in fact it is an OrderedDict
        validator=attr.validators.deep_mapping(
            attr.validators.instance_of(str),
            attr.validators.deep_iterable(
                attr.validators.instance_of(str),
                attr.validators.instance_of(list)),
            attr.validators.instance_of(OrderedDict)),
        default=SOUND_MODE_MAPPING,
        init=False)
    _sound_mode_map_rev: Dict[str, str] = attr.ib(
        validator=attr.validators.deep_mapping(
            attr.validators.instance_of(str),
            attr.validators.instance_of(str),
            attr.validators.instance_of(dict)),
        default=attr.Factory(sound_mode_rev_map_factory, takes_self=True),
        init=False)

    # Update tags for attributes
    # AppCommand.xml interface
    appcommand_attrs = {
        AppCommands.GetSurroundModeStatus: None}
    # Status.xml interface
    status_xml_attrs_01 = {
        "_sound_mode_raw": "./selectSurround/value"}
    status_xml_attrs_02 = {
        "_sound_mode_raw": "./SurrMode/value"}

    def setup(self) -> None:
        """Ensure that the instance is initialized."""
        # Add tags for a potential AppCommand.xml update
        for tag in self.appcommand_attrs:
            self._device.api.add_appcommand_update_tag(tag)

        # Soundmode is always available for AVR-X and AVR-X-2016 receivers
        # For AVR receiver it will be tested druing the first update
        if self._device.receiver in [AVR_X, AVR_X_2016]:
            self._support_sound_mode = True

        self._is_setup = True

    async def async_update(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None) -> None:
        """Update sound mode asynchronously."""
        # Ensure instance is setup before updating
        if self._is_setup is False:
            self.setup()

        # Update state
        await self.async_update_sound_mode(
            global_update=global_update, cache_id=cache_id)

    async def async_update_sound_mode(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None):
        """Update volume status of device."""
        if self._device.use_avr_2016_update is True:
            await self.async_update_attrs_appcommand(
                self.appcommand_attrs, global_update=global_update,
                cache_id=cache_id)
        elif self._device.use_avr_2016_update is False:
            urls = [self._device.urls.status, self._device.urls.mainzone]
            if self._support_sound_mode is False:
                return
            # There are two different options of sound mode tags
            try:
                await self.async_update_attrs_status_xml(
                    self.status_xml_attrs_01, urls, cache_id=cache_id)
            except AvrProcessingError:
                try:
                    await self.async_update_attrs_status_xml(
                        self.status_xml_attrs_02, urls, cache_id=cache_id)
                except AvrProcessingError:
                    _LOGGER.info("Sound mode not supported")
                    self._support_sound_mode = False
                    return
            self._support_sound_mode = True
        else:
            raise AvrProcessingError(
                "Device is not setup correctly, update method not set")

    def match_sound_mode(self, sound_mode_raw: str) -> Optional[str]:
        """Match the raw_sound_mode to its corresponding sound_mode."""
        if self._sound_mode_raw is None:
            return None
        try:
            sound_mode = self._sound_mode_map_rev[sound_mode_raw.upper()]
            return sound_mode
        except KeyError:
            smr_up = sound_mode_raw.upper()
            self._sound_mode_map[smr_up] = [smr_up]
            self._sound_mode_map_rev = sound_mode_rev_map_factory(self)
            _LOGGER.warning("Not able to match sound mode: '%s', "
                            "returning raw sound mode.", smr_up)
        return sound_mode_raw

    async def _async_set_all_zone_stereo(self, zst_on: bool) -> None:
        """
        Set All Zone Stereo option on the device.

        Calls command to activate/deactivate the mode
        """
        command_url = self._device.urls.command_set_all_zone_stereo
        if zst_on:
            command_url += "ZST ON"
        else:
            command_url += "ZST OFF"

        await self._device.api.async_get_command(command_url)

    ##############
    # Properties #
    ##############
    @property
    def support_sound_mode(self) -> Optional[bool]:
        """Return True if sound mode supported."""
        return self._support_sound_mode

    @property
    def sound_mode(self) -> Optional[str]:
        """Return the matched current sound mode as a string."""
        sound_mode_matched = self.match_sound_mode(self._sound_mode_raw)
        return sound_mode_matched

    @property
    def sound_mode_list(self) -> None:
        """Return a list of available sound modes as string."""
        return list(self._sound_mode_map.keys())

    @property
    def sound_mode_map(self) -> Dict[str, str]:  # returns an OrderedDict
        """Return a dict of available sound modes with their mapping values."""
        return deepcopy(self._sound_mode_map)

    @property
    def sound_mode_map_rev(self) -> Dict[str, str]:
        """Return a dict to map each sound_mode_raw to matching sound_mode."""
        return deepcopy(self._sound_mode_map_rev)

    @property
    def sound_mode_raw(self) -> Optional[str]:
        """Return the current sound mode as string as received from the AVR."""
        return self._sound_mode_raw

    ##########
    # Setter #
    ##########
    async def async_set_sound_mode(self, sound_mode: str) -> None:
        """
        Set sound_mode of device.

        Valid values depend on the device and should be taken from
        "sound_mode_list".
        """
        if sound_mode == ALL_ZONE_STEREO:
            await self._async_set_all_zone_stereo(True)
            return

        if self.sound_mode == ALL_ZONE_STEREO:
            self._async_set_all_zone_stereo(False)
        # For selection of sound mode other names then at receiving sound modes
        # have to be used
        # Therefore source mapping is needed to get sound_mode
        # Create command URL and send command via HTTP GET
        command_url = self._device.urls.command_sel_sound_mode + sound_mode
        # sent command
        await self._device.api.async_get_command(command_url)


def sound_mode_factory(instance: DenonAVRFoundation) -> DenonAVRSoundMode:
    """Create DenonAVRSoundMode at receiver instances."""
    # pylint: disable=protected-access
    new = DenonAVRSoundMode(device=instance._device)
    return new



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module covers some basic automated tests of Denon AVR receivers.

:copyright: (c) 2016 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import httpx
import pytest
from pytest_httpx import HTTPXMock, to_response

import denonavr

FAKE_IP = "10.0.0.0"

NO_ZONES = None
ZONE2 = {"Zone2": None}
ZONE3 = {"Zone3": None}
ZONE2_ZONE3 = {"Zone2": None, "Zone3": None}

TESTING_RECEIVERS = {
    "AVR-X4100W": (NO_ZONES, denonavr.const.AVR_X),
    "AVR-2312CI": (NO_ZONES, denonavr.const.AVR),
    "AVR-1912": (NO_ZONES, denonavr.const.AVR),
    "AVR-3311CI": (NO_ZONES, denonavr.const.AVR),
    "M-RC610": (NO_ZONES, denonavr.const.AVR_X),
    "AVR-X2100W-2": (NO_ZONES, denonavr.const.AVR_X),
    "AVR-X2000": (ZONE2_ZONE3, denonavr.const.AVR_X),
    "AVR-X2000-2": (NO_ZONES, denonavr.const.AVR_X),
    "SR5008": (NO_ZONES, denonavr.const.AVR_X),
    "M-CR603": (NO_ZONES, denonavr.const.AVR),
    "NR1604": (ZONE2_ZONE3, denonavr.const.AVR_X),
    "AVR-4810": (NO_ZONES, denonavr.const.AVR),
    "AVR-3312": (NO_ZONES, denonavr.const.AVR),
    "NR1609": (ZONE2, denonavr.const.AVR_X_2016),
    "AVC-8500H": (ZONE2_ZONE3, denonavr.const.AVR_X_2016),
    "AVR-X4300H": (ZONE2_ZONE3, denonavr.const.AVR_X_2016),
    "AVR-X1100W": (ZONE2, denonavr.const.AVR_X),
    "SR6012": (ZONE2, denonavr.const.AVR_X_2016),
    }

APPCOMMAND_URL = "/goform/AppCommand.xml"
STATUS_URL = "/goform/formMainZone_MainZoneXmlStatus.xml"
STATUS_Z2_URL = "/goform/formZone2_Zone2XmlStatus.xml"
STATUS_Z3_URL = "/goform/formZone3_Zone3XmlStatus.xml"
MAINZONE_URL = "/goform/formMainZone_MainZoneXml.xml"
DEVICEINFO_URL = "/goform/Deviceinfo.xml"
NETAUDIOSTATUS_URL = "/goform/formNetAudio_StatusXml.xml"
TUNERSTATUS_URL = "/goform/formTuner_TunerXml.xml"
HDTUNERSTATUS_URL = "/goform/formTuner_HdXml.xml"
DESCRIPTION_URL1 = "/description.xml"
DESCRIPTION_URL2 = "/upnp/desc/aios_device/aios_device.xml"


def get_sample_content(filename):
    """Return sample content form file."""
    with open("tests/xml/{filename}".format(filename=filename),
              encoding="utf-8") as file:
        return file.read()


class TestMainFunctions:
    """Test case for main functions of Denon AVR."""

    testing_receiver = None
    denon = None

    def custom_matcher(self, request: httpx.Request, *args, **kwargs):
        """Match URLs to sample files."""
        port_suffix = ""

        if request.url.port == 8080:
            port_suffix = "-8080"

        try:
            if request.url.path == STATUS_URL:
                content = get_sample_content(
                    "{receiver}-formMainZone_MainZoneXmlStatus{port}"
                    ".xml".format(
                        receiver=self.testing_receiver, port=port_suffix))
            elif request.url.path == STATUS_Z2_URL:
                content = get_sample_content(
                    "{receiver}-formZone2_Zone2XmlStatus{port}.xml".format(
                        receiver=self.testing_receiver, port=port_suffix))
            elif request.url.path == STATUS_Z3_URL:
                content = get_sample_content(
                    "{receiver}-formZone3_Zone3XmlStatus{port}.xml".format(
                        receiver=self.testing_receiver, port=port_suffix))
            elif request.url.path == MAINZONE_URL:
                content = get_sample_content(
                    "{receiver}-formMainZone_MainZoneXml{port}.xml".format(
                        receiver=self.testing_receiver, port=port_suffix))
            elif request.url.path == DEVICEINFO_URL:
                content = get_sample_content(
                    "{receiver}-Deviceinfo{port}.xml".format(
                        receiver=self.testing_receiver, port=port_suffix))
            elif request.url.path == NETAUDIOSTATUS_URL:
                content = get_sample_content(
                    "{receiver}-formNetAudio_StatusXml{port}.xml".format(
                        receiver=self.testing_receiver, port=port_suffix))
            elif request.url.path == TUNERSTATUS_URL:
                content = get_sample_content(
                    "{receiver}-formTuner_TunerXml{port}.xml".format(
                        receiver=self.testing_receiver, port=port_suffix))
            elif request.url.path == HDTUNERSTATUS_URL:
                content = get_sample_content(
                    "{receiver}-formTuner_HdXml{port}.xml".format(
                        receiver=self.testing_receiver, port=port_suffix))
            elif request.url.path == APPCOMMAND_URL:
                content_str = request.read().decode("utf-8")
                if "GetFriendlyName" in content_str:
                    ep_suffix = "-setup"
                else:
                    ep_suffix = "-update"
                content = get_sample_content(
                    "{receiver}-AppCommand{ep}{port}.xml".format(
                        receiver=self.testing_receiver,
                        port=port_suffix,
                        ep=ep_suffix))
            elif request.url.path in [DESCRIPTION_URL1, DESCRIPTION_URL2]:
                content = get_sample_content("AVR-X1600H_upnp.xml")
            else:
                content = "DATA"
        except FileNotFoundError:
            content = "Error 403: Forbidden\nAccess Forbidden"
            status_code = 403
        else:
            status_code = 200

        resp = to_response(status_code=status_code, data=content)

        return resp

    @pytest.mark.asyncio
    async def test_receiver_type(self, httpx_mock: HTTPXMock):
        """Check that receiver type is determined correctly."""
        httpx_mock.add_callback(self.custom_matcher)
        for receiver, spec in TESTING_RECEIVERS.items():
            print("Receiver: {}".format(receiver))
            # Switch receiver and update to load new sample files
            self.testing_receiver = receiver
            self.denon = denonavr.DenonAVR(FAKE_IP, add_zones=spec[0])
            await self.denon.async_update()
            assert self.denon.receiver_type == spec[1].type, (
                "Receiver type is {} not {} for receiver {}".format(
                    self.denon.receiver_type, spec[1].type, receiver))
            assert self.denon.receiver_port == spec[1].port, (
                "Receiver port is {} not {} for receiver {}".format(
                    self.denon.receiver_port, spec[1].port, receiver))

    @pytest.mark.asyncio
    async def test_input_func_switch(self, httpx_mock: HTTPXMock):
        """Switch through all input functions of all tested receivers."""
        httpx_mock.add_callback(self.custom_matcher)
        for receiver, spec in TESTING_RECEIVERS.items():
            # Switch receiver and update to load new sample files
            self.testing_receiver = receiver
            self.denon = denonavr.DenonAVR(FAKE_IP, add_zones=spec[0])
            # Switch through all functions and check if successful
            for name, zone in self.denon.zones.items():
                print("Receiver: {}, Zone: {}".format(receiver, name))
                await self.denon.zones[name].async_update()
                assert len(zone.input_func_list) > 0
                for input_func in zone.input_func_list:
                    await self.denon.zones[name].async_set_input_func(
                        input_func)

    @pytest.mark.asyncio
    async def test_attributes_not_none(self, httpx_mock: HTTPXMock):
        """Check that certain attributes are not None."""
        httpx_mock.add_callback(self.custom_matcher)
        for receiver, spec in TESTING_RECEIVERS.items():
            print("Receiver: {}".format(receiver))
            # Switch receiver and update to load new sample files
            self.testing_receiver = receiver
            self.denon = denonavr.DenonAVR(FAKE_IP, add_zones=spec[0])
            await self.denon.async_update()
            assert self.denon.power is not None, (
                "Power status is None for receiver {}".format(receiver))
            assert self.denon.state is not None, (
                "State is None for receiver {}".format(receiver))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module inherits constants for Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import re

from collections import namedtuple, OrderedDict

import attr

# Attr constants
DENON_ATTR_SETATTR = [attr.setters.validate, attr.setters.convert]

# Types for constants
ReceiverType = namedtuple("ReceiverType", ["type", "port"])
DescriptionType = namedtuple("DescriptionType", ["port", "url"])
ReceiverURLs = namedtuple(
    "ReceiverURLs", [
        "appcommand", "appcommand0300", "status", "mainzone", "deviceinfo",
        "netaudiostatus", "tunerstatus", "hdtunerstatus", "command_sel_src",
        "command_fav_src", "command_power_on", "command_power_standby",
        "command_volume_up", "command_volume_down", "command_set_volume",
        "command_mute_on", "command_mute_off", "command_sel_sound_mode",
        "command_netaudio_post", "command_set_all_zone_stereo",
        "command_pause", "command_play"])

# AVR-X search patterns
DEVICEINFO_AVR_X_PATTERN = re.compile(
    r"(.*AV(C|R)-(X|S).*|.*SR500[6-9]|.*SR60(07|08|09|10|11|12|13)|."
    r"*SR70(07|08|09|10|11|12|13)|.*SR501[3-4]|.*NR1604|.*NR1710)")

DEVICEINFO_COMMAPI_PATTERN = re.compile(r"(0210|0220|0250|0300|0301)")

# Receiver types
AVR_NAME = "avr"
AVR_X_NAME = "avr-x"
AVR_X_2016_NAME = "avr-x-2016"

AVR = ReceiverType(type=AVR_NAME, port=80)
AVR_X = ReceiverType(type=AVR_X_NAME, port=80)
AVR_X_2016 = ReceiverType(type=AVR_X_2016_NAME, port=8080)
VALID_RECEIVER_TYPES = (AVR, AVR_X, AVR_X_2016)

DESCRIPTION_TYPES = {
    AVR_NAME: DescriptionType(port=8080, url="/description.xml"),
    AVR_X_NAME: DescriptionType(port=8080, url="/description.xml"),
    AVR_X_2016_NAME: DescriptionType(
        port=60006, url="/upnp/desc/aios_device/aios_device.xml")}

SOURCE_MAPPING = {
    "TV AUDIO": "TV", "iPod/USB": "USB/IPOD", "Bluetooth": "BT",
    "Blu-ray": "BD", "CBL/SAT": "SAT/CBL", "NETWORK": "NET",
    "Media Player": "MPLAY", "AUX": "AUX1", "Tuner": "TUNER",
    "FM": "TUNER", "SpotifyConnect": "Spotify Connect"}

CHANGE_INPUT_MAPPING = {"Internet Radio": "IRP", "Online Music": "NET",
                        "Media Server": "SERVER", "Spotify": "SPOTIFY",
                        "Flickr": "FLICKR", "Favorites": "FAVORITES"}

# Sound modes
ALL_ZONE_STEREO = "ALL ZONE STEREO"

SOUND_MODE_MAPPING = OrderedDict(
    [("MUSIC", ["PLII MUSIC", "DTS NEO:6 MUSIC", "DOLBY D +NEO:X M",
                "DTS NEO:X MUSIC", "DOLBY PL2 MUSIC", "PLIIX MUSIC"]),
     ("MOVIE", ["PLII MOVIE", "PLII CINEMA", "DTS NEO:X CINEMA",
                "DTS NEO:6 CINEMA", "DOLBY D +NEO:X C",
                "PLIIX CINEMA", "DOLBY PLII MOVIE", "MULTI IN + VIRTUAL:X"]),
     ("GAME", ["PLII GAME", "DOLBY D +NEO:X G"]),
     ("AUTO", ["None"]),
     ("STANDARD", ["None2"]),
     ("VIRTUAL", ["VIRTUAL"]),
     ("MATRIX", ["MATRIX"]),
     ("ROCK ARENA", ["ROCK ARENA"]),
     ("JAZZ CLUB", ["JAZZ CLUB"]),
     ("VIDEO GAME", ["VIDEO GAME"]),
     ("MONO MOVIE", ["MONO MOVIE"]),
     ("DIRECT", ["DIRECT"]),
     ("PURE DIRECT", ["PURE_DIRECT", "PURE DIRECT"]),
     ("DOLBY DIGITAL", ["DOLBY DIGITAL", "DOLBY D + DOLBY SURROUND",
                        "DOLBY DIGITAL +", "STANDARD(DOLBY)", "DOLBY SURROUND",
                        "DOLBY D + +DOLBY SURROUND", "NEURAL", "DOLBY HD",
                        "DOLBY HD + DOLBY SURROUND",
                        "MULTI IN + NEURAL:X", "MULTI IN + DOLBY SURROUND",
                        "DOLBY D + NEURAL:X", "DOLBY DIGITAL + NEURAL:X",
                        "DOLBY DIGITAL + + NEURAL:X", "DOLBY ATMOS",
                        "DOLBY AUDIO - DOLBY SURROUND", "DOLBY TRUEHD",
                        "DOLBY AUDIO - DOLBY DIGITAL PLUS",
                        "DOLBY AUDIO - TRUEHD + DSUR",
                        "DOLBY AUDIO - DOLBY TRUEHD",
                        "DOLBY AUDIO - TRUEHD + NEURAL:X",
                        "DOLBY AUDIO - DD + DSUR",
                        "DOLBY AUDIO - DD+   + NEURAL:X",
                        "DOLBY AUDIO - DD+   + DSUR",
                        "DOLBY AUDIO - DOLBY DIGITAL"]),
     ("DTS SURROUND", ["DTS SURROUND", "DTS NEURAL:X", "STANDARD(DTS)",
                       "DTS + NEURAL:X", "MULTI CH IN", "DTS-HD MSTR",
                       "DTS VIRTUAL:X", "DTS-HD + NEURAL:X", "DTS-HD",
                       "DTS + VIRTUAL:X", "DTS + DOLBY SURROUND",
                       "DTS-HD + DOLBY SURROUND", "DTS:X MSTR"]),
     ("AURO3D", ["AURO-3D", "AURO-2D SURROUND"]),
     ("MCH STEREO", ["MULTI CH STEREO", "MULTI_CH_STEREO", "MCH STEREO",
                     "MULTI CH IN 7.1"]),
     ("STEREO", ["STEREO"]),
     (ALL_ZONE_STEREO, ["ALL ZONE STEREO"])])

# Receiver sources
PLAYING_SOURCES = (
    "Online Music", "Media Server", "iPod/USB", "Bluetooth",
    "Internet Radio", "Favorites", "SpotifyConnect", "Flickr",
    "TUNER", "NET/USB", "HDRADIO", "Music Server", "NETWORK", "NET")
NETAUDIO_SOURCES = (
    "Online Music", "Media Server", "iPod/USB", "Bluetooth",
    "Internet Radio", "Favorites", "SpotifyConnect", "Flickr",
    "NET/USB", "Music Server", "NETWORK", "NET")

# Image URLs
STATIC_ALBUM_URL = "http://{host}:{port}/img/album%20art_S.png"
ALBUM_COVERS_URL = "http://{host}:{port}/NetAudio/art.asp-jpg?{time}"

# General URLs
APPCOMMAND_URL = "/goform/AppCommand.xml"
APPCOMMAND0300_URL = "/goform/AppCommand0300.xml"
DEVICEINFO_URL = "/goform/Deviceinfo.xml"
NETAUDIOSTATUS_URL = "/goform/formNetAudio_StatusXml.xml"
TUNERSTATUS_URL = "/goform/formTuner_TunerXml.xml"
HDTUNERSTATUS_URL = "/goform/formTuner_HdXml.xml"
COMMAND_NETAUDIO_POST_URL = "/NetAudio/index.put.asp"
COMMAND_PAUSE = "/goform/formiPhoneAppDirect.xml?NS9B"
COMMAND_PLAY = "/goform/formiPhoneAppDirect.xml?NS9A"


# Main Zone URLs
STATUS_URL = "/goform/formMainZone_MainZoneXmlStatus.xml"
MAINZONE_URL = "/goform/formMainZone_MainZoneXml.xml"
COMMAND_SEL_SRC_URL = "/goform/formiPhoneAppDirect.xml?SI"
COMMAND_FAV_SRC_URL = "/goform/formiPhoneAppDirect.xml?ZM"
COMMAND_POWER_ON_URL = "/goform/formiPhoneAppPower.xml?1+PowerOn"
COMMAND_POWER_STANDBY_URL = "/goform/formiPhoneAppPower.xml?1+PowerStandby"
COMMAND_VOLUME_UP_URL = "/goform/formiPhoneAppDirect.xml?MVUP"
COMMAND_VOLUME_DOWN_URL = "/goform/formiPhoneAppDirect.xml?MVDOWN"
COMMAND_SET_VOLUME_URL = "/goform/formiPhoneAppVolume.xml?1+{volume:.1f}"
COMMAND_MUTE_ON_URL = "/goform/formiPhoneAppMute.xml?1+MuteOn"
COMMAND_MUTE_OFF_URL = "/goform/formiPhoneAppMute.xml?1+MuteOff"
COMMAND_SEL_SM_URL = "/goform/formiPhoneAppDirect.xml?MS"
COMMAND_SET_ZST_URL = "/goform/formiPhoneAppDirect.xml?MN"

# Zone 2 URLs
STATUS_Z2_URL = "/goform/formZone2_Zone2XmlStatus.xml"
MAINZONE_Z2_URL = None
COMMAND_SEL_SRC_Z2_URL = "/goform/formiPhoneAppDirect.xml?Z2"
COMMAND_FAV_SRC_Z2_URL = "/goform/formiPhoneAppDirect.xml?Z2"
COMMAND_POWER_ON_Z2_URL = "/goform/formiPhoneAppPower.xml?2+PowerOn"
COMMAND_POWER_STANDBY_Z2_URL = "/goform/formiPhoneAppPower.xml?2+PowerStandby"
COMMAND_VOLUME_UP_Z2_URL = "/goform/formiPhoneAppDirect.xml?Z2UP"
COMMAND_VOLUME_DOWN_Z2_URL = "/goform/formiPhoneAppDirect.xml?Z2DOWN"
COMMAND_SET_VOLUME_Z2_URL = "/goform/formiPhoneAppVolume.xml?2+{volume:.1f}"
COMMAND_MUTE_ON_Z2_URL = "/goform/formiPhoneAppMute.xml?2+MuteOn"
COMMAND_MUTE_OFF_Z2_URL = "/goform/formiPhoneAppMute.xml?2+MuteOff"

# Zone 3 URLs
STATUS_Z3_URL = "/goform/formZone3_Zone3XmlStatus.xml"
MAINZONE_Z3_URL = None
COMMAND_SEL_SRC_Z3_URL = "/goform/formiPhoneAppDirect.xml?Z3"
COMMAND_FAV_SRC_Z3_URL = "/goform/formiPhoneAppDirect.xml?Z3"
COMMAND_POWER_ON_Z3_URL = "/goform/formiPhoneAppPower.xml?3+PowerOn"
COMMAND_POWER_STANDBY_Z3_URL = "/goform/formiPhoneAppPower.xml?3+PowerStandby"
COMMAND_VOLUME_UP_Z3_URL = "/goform/formiPhoneAppDirect.xml?Z3UP"
COMMAND_VOLUME_DOWN_Z3_URL = "/goform/formiPhoneAppDirect.xml?Z3DOWN"
COMMAND_SET_VOLUME_Z3_URL = "/goform/formiPhoneAppVolume.xml?3+{volume:.1f}"
COMMAND_MUTE_ON_Z3_URL = "/goform/formiPhoneAppMute.xml?3+MuteOn"
COMMAND_MUTE_OFF_Z3_URL = "/goform/formiPhoneAppMute.xml?3+MuteOff"

DENONAVR_URLS = ReceiverURLs(
    appcommand=APPCOMMAND_URL,
    appcommand0300=APPCOMMAND0300_URL,
    status=STATUS_URL,
    mainzone=MAINZONE_URL,
    deviceinfo=DEVICEINFO_URL,
    netaudiostatus=NETAUDIOSTATUS_URL,
    tunerstatus=TUNERSTATUS_URL,
    hdtunerstatus=HDTUNERSTATUS_URL,
    command_sel_src=COMMAND_SEL_SRC_URL,
    command_fav_src=COMMAND_FAV_SRC_URL,
    command_power_on=COMMAND_POWER_ON_URL,
    command_power_standby=COMMAND_POWER_STANDBY_URL,
    command_volume_up=COMMAND_VOLUME_UP_URL,
    command_volume_down=COMMAND_VOLUME_DOWN_URL,
    command_set_volume=COMMAND_SET_VOLUME_URL,
    command_mute_on=COMMAND_MUTE_ON_URL,
    command_mute_off=COMMAND_MUTE_OFF_URL,
    command_sel_sound_mode=COMMAND_SEL_SM_URL,
    command_netaudio_post=COMMAND_NETAUDIO_POST_URL,
    command_set_all_zone_stereo=COMMAND_SET_ZST_URL,
    command_pause=COMMAND_PAUSE,
    command_play=COMMAND_PLAY)

ZONE2_URLS = ReceiverURLs(
    appcommand=APPCOMMAND_URL,
    appcommand0300=APPCOMMAND0300_URL,
    status=STATUS_Z2_URL,
    mainzone=MAINZONE_Z2_URL,
    deviceinfo=DEVICEINFO_URL,
    netaudiostatus=NETAUDIOSTATUS_URL,
    tunerstatus=TUNERSTATUS_URL,
    hdtunerstatus=HDTUNERSTATUS_URL,
    command_sel_src=COMMAND_SEL_SRC_Z2_URL,
    command_fav_src=COMMAND_FAV_SRC_Z2_URL,
    command_power_on=COMMAND_POWER_ON_Z2_URL,
    command_power_standby=COMMAND_POWER_STANDBY_Z2_URL,
    command_volume_up=COMMAND_VOLUME_UP_Z2_URL,
    command_volume_down=COMMAND_VOLUME_DOWN_Z2_URL,
    command_set_volume=COMMAND_SET_VOLUME_Z2_URL,
    command_mute_on=COMMAND_MUTE_ON_Z2_URL,
    command_mute_off=COMMAND_MUTE_OFF_Z2_URL,
    command_sel_sound_mode=COMMAND_SEL_SM_URL,
    command_netaudio_post=COMMAND_NETAUDIO_POST_URL,
    command_set_all_zone_stereo=COMMAND_SET_ZST_URL,
    command_pause=COMMAND_PAUSE,
    command_play=COMMAND_PLAY)

ZONE3_URLS = ReceiverURLs(
    appcommand=APPCOMMAND_URL,
    appcommand0300=APPCOMMAND0300_URL,
    status=STATUS_Z3_URL,
    mainzone=MAINZONE_Z3_URL,
    deviceinfo=DEVICEINFO_URL,
    netaudiostatus=NETAUDIOSTATUS_URL,
    tunerstatus=TUNERSTATUS_URL,
    hdtunerstatus=HDTUNERSTATUS_URL,
    command_sel_src=COMMAND_SEL_SRC_Z3_URL,
    command_fav_src=COMMAND_FAV_SRC_Z3_URL,
    command_power_on=COMMAND_POWER_ON_Z3_URL,
    command_power_standby=COMMAND_POWER_STANDBY_Z3_URL,
    command_volume_up=COMMAND_VOLUME_UP_Z3_URL,
    command_volume_down=COMMAND_VOLUME_DOWN_Z3_URL,
    command_set_volume=COMMAND_SET_VOLUME_Z3_URL,
    command_mute_on=COMMAND_MUTE_ON_Z3_URL,
    command_mute_off=COMMAND_MUTE_OFF_Z3_URL,
    command_sel_sound_mode=COMMAND_SEL_SM_URL,
    command_netaudio_post=COMMAND_NETAUDIO_POST_URL,
    command_set_all_zone_stereo=COMMAND_SET_ZST_URL,
    command_pause=COMMAND_PAUSE,
    command_play=COMMAND_PLAY)

# States
POWER_ON = "ON"
POWER_OFF = "OFF"
POWER_STANDBY = "STANDBY"
STATE_ON = "on"
STATE_OFF = "off"
STATE_PLAYING = "playing"
STATE_PAUSED = "paused"

# Zones
MAIN_ZONE = "Main"
ZONE2 = "Zone2"
ZONE3 = "Zone3"
VALID_ZONES = [MAIN_ZONE, ZONE2, ZONE3]

# Setup additional zones
NO_ZONES = None
ZONE2_NAME = {ZONE2: None}
ZONE3_NAME = {ZONE3: None}
ZONE2_ZONE3_NAME = {ZONE2: None, ZONE3: None}

# AppCommand related
APPCOMMAND_CMD_TEXT = "cmd_text"
APPCOMMAND_NAME = "name"

# Audyssey parameter
MULTI_EQ_MAP = {"0": "Off", "1": "Flat", "2": "L/R Bypass", "3": "Reference"}
MULTI_EQ_MAP_LABELS = {(value, key) for key, value in MULTI_EQ_MAP.items()}

REF_LVL_OFFSET_MAP = {"0": "0dB", "1": "+5dB", "2": "+10dB", "3": "+15dB"}
REF_LVL_OFFSET_MAP_LABELS = {
    (value, key) for key, value in REF_LVL_OFFSET_MAP.items()}

DYNAMIC_VOLUME_MAP = {"0": "Off", "1": "Light", "2": "Medium", "3": "Heavy"}
DYNAMIC_VOLUME_MAP_LABELS = {
    (value, key) for key, value in DYNAMIC_VOLUME_MAP.items()}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the REST API to Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import logging
import xml.etree.ElementTree as ET

from io import BytesIO
from typing import Dict, Hashable, Optional, Tuple

import attr
import httpx

from async_lru import alru_cache
from defusedxml.ElementTree import fromstring

from .appcommand import AppCommandCmd
from .decorators import (
    async_handle_receiver_exceptions,
    set_cache_id)
from .exceptions import AvrInvalidResponseError
from .const import (
    APPCOMMAND_CMD_TEXT, APPCOMMAND_NAME, APPCOMMAND_URL, APPCOMMAND0300_URL,
    DENON_ATTR_SETATTR)

_LOGGER = logging.getLogger(__name__)


@attr.s(auto_attribs=True, hash=False, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRApi:
    """Perform API calls to Denon AVR REST interface."""

    host: str = attr.ib(converter=str, default="localhost")
    port: int = attr.ib(converter=int, default=80)
    timeout: httpx.Timeout = attr.ib(
        validator=attr.validators.instance_of(httpx.Timeout),
        default=httpx.Timeout(2.0, read=15.0))
    _appcommand_update_tags: Tuple[AppCommandCmd] = attr.ib(
        validator=attr.validators.deep_iterable(
            attr.validators.instance_of(AppCommandCmd),
            attr.validators.instance_of(tuple)),
        default=attr.Factory(tuple))
    _appcommand0300_update_tags: Tuple[AppCommandCmd] = attr.ib(
        validator=attr.validators.deep_iterable(
            attr.validators.instance_of(AppCommandCmd),
            attr.validators.instance_of(tuple)),
        default=attr.Factory(tuple))
    async_client: httpx.AsyncClient = attr.ib(
        validator=attr.validators.in_((httpx.AsyncClient,)),
        default=httpx.AsyncClient, init=False)

    def __hash__(self) -> int:
        """
        Hash the class in a custom way that caching works.

        It should react on changes of host and port.
        """
        return hash((id(self), self.host, self.port))

    @async_handle_receiver_exceptions
    async def async_get(
            self,
            request: str,
            port: Optional[int] = None) -> httpx.Response:
        """Call GET endpoint of Denon AVR receiver asynchronously."""
        # Use default port of the receiver if no different port is specified
        port = port if port is not None else self.port

        endpoint = "http://{host}:{port}{request}".format(
            host=self.host, port=port, request=request)

        async with self.async_client() as client:
            res = await client.get(endpoint, timeout=self.timeout)
            res.raise_for_status()

        return res

    @async_handle_receiver_exceptions
    async def async_post(
            self,
            request: str,
            content: Optional[bytes] = None,
            data: Optional[Dict] = None,
            port: Optional[int] = None) -> httpx.Response:
        """Call POST endpoint of Denon AVR receiver asynchronously."""
        # Use default port of the receiver if no different port is specified
        port = port if port is not None else self.port

        endpoint = "http://{host}:{port}{request}".format(
            host=self.host, port=port, request=request)

        async with self.async_client() as client:
            res = await client.post(
                endpoint, content=content, data=data, timeout=self.timeout)
            res.raise_for_status()

        return res

    @async_handle_receiver_exceptions
    async def async_get_command(self, request: str) -> str:
        """Send HTTP GET command to Denon AVR receiver asynchronously."""
        # HTTP GET to endpoint
        res = await self.async_get(request)
        # Return text
        return res.text

    @set_cache_id
    @alru_cache(maxsize=32, cache_exceptions=False)
    @async_handle_receiver_exceptions
    async def async_get_xml(
            self,
            request: str,
            cache_id: Hashable = None
            ) -> ET.Element:
        """Return XML data from HTTP GET endpoint asynchronously."""
        # HTTP GET to endpoint
        res = await self.async_get(request)
        # create ElementTree
        xml_root = fromstring(res.text)
        # Check validity of XML
        self.check_xml_validity(request, xml_root)
        # Return ElementTree element
        return xml_root

    @set_cache_id
    @alru_cache(maxsize=32, cache_exceptions=False)
    @async_handle_receiver_exceptions
    async def async_post_appcommand(
            self,
            request: str,
            cmds: Tuple[AppCommandCmd],
            cache_id: Hashable = None
            ) -> ET.Element:
        """Return XML from Appcommand(0300) endpoint asynchronously."""
        # Prepare XML body for POST call
        content = self.prepare_appcommand_body(cmds)
        _LOGGER.debug("Content for %s endpoint: %s", request, content)
        # HTTP POST to endpoint
        res = await self.async_post(request, content=content)
        # create ElementTree
        xml_root = fromstring(res.text)
        # Check validity of XML
        self.check_xml_validity(request, xml_root)
        # Add query tags to result
        xml_root = self.add_query_tags_to_result(request, cmds, xml_root)
        # Return ElementTree element
        return xml_root

    def add_appcommand_update_tag(self, tag: AppCommandCmd) -> None:
        """Add appcommand tag for full update."""
        if tag.cmd_id != "1":
            raise ValueError("cmd_id is {} but must be 1".format(tag.cmd_id))

        # Remove response pattern from tag because it is not relevant for query
        tag = attr.evolve(tag, response_pattern=tuple())

        if tag not in self._appcommand_update_tags:
            _LOGGER.debug("Add tag %s to AppCommand update tuple", tag)
            self._appcommand_update_tags = (*self._appcommand_update_tags, tag)

    def add_appcommand0300_update_tag(self, tag: AppCommandCmd) -> None:
        """Add appcommand0300 tag for full update."""
        if tag.cmd_id != "3":
            raise ValueError("cmd_id is {} but must be 3".format(tag.cmd_id))

        # Remove response pattern from tag because it is not relevant for query
        tag = attr.evolve(tag, response_pattern=tuple())

        if tag not in self._appcommand0300_update_tags:
            _LOGGER.debug("Add tag %s to AppCommand0300 update tuple", tag)
            self._appcommand0300_update_tags = (
                *self._appcommand0300_update_tags, tag)

    async def async_get_global_appcommand(
            self,
            appcommand0300: bool = False,
            cache_id: Optional[Hashable] = None) -> ET.Element:
        """Get XML structure for full AppCommand update."""
        if appcommand0300:
            tags = self._appcommand0300_update_tags
            url = APPCOMMAND0300_URL
        else:
            tags = self._appcommand_update_tags
            url = APPCOMMAND_URL

        res = await self.async_post_appcommand(url, tags, cache_id=cache_id)
        return res

    @staticmethod
    def add_query_tags_to_result(
            request: str,
            cmd_list: Tuple[AppCommandCmd],
            xml_root: ET.Element
            ) -> ET.Element:
        """
        Add query tags to corresponding elements of result XML.

        This is used to identitfy the result tags.
        """
        if len(cmd_list) != len(xml_root):
            raise AvrInvalidResponseError(
                "Invalid length of response XML. Query has {} elements, "
                "response {}".format(len(cmd_list), len(xml_root)), request)

        for i, child in enumerate(xml_root):
            if child.tag not in ["cmd", "error"]:
                raise AvrInvalidResponseError(
                    "Returned document contains a tag other than \"cmd\" and "
                    "\"error\": {}".format(child.tag), request)
            # Find corresponding attributes from request XML if set and add
            # tag to cmd element
            if cmd_list[i].cmd_text is not None:
                child.set(APPCOMMAND_CMD_TEXT, cmd_list[i].cmd_text)
            if cmd_list[i].name is not None:
                child.set(APPCOMMAND_NAME, cmd_list[i].name)

        return xml_root

    @staticmethod
    def check_xml_validity(request: str, xml_root: ET.Element) -> None:
        """Check if there is a valid Denon XML and not a HTML page."""
        if xml_root.tag == "html":
            raise AvrInvalidResponseError(
                "Returned document contains HTML", request)

    @staticmethod
    def prepare_appcommand_body(cmd_list: Tuple[AppCommandCmd]) -> bytes:
        """Prepare HTTP POST body to AppCommand(0300).xml end point."""
        # Buffer XML body as binary IO
        body = BytesIO()

        # Denon AppCommand.xml acts weird. It returns an error when the tx
        # element consists of more than 5 cmd elements, but it accepts
        # multiple XML root elements
        chunks = [cmd_list[i:i+5] for i in range(
            0, len(cmd_list), 5)]

        for i, chunk in enumerate(chunks):
            # Prepare POST XML body for AppCommand.xml
            xml_root = ET.Element("tx")

            for cmd in chunk:
                # Append tags for each element
                cmd_element = ET.Element("cmd")
                cmd_element.set("id", cmd.cmd_id)

                # Optional cmd text
                cmd_element.text = cmd.cmd_text

                # Optional name tag
                if cmd.name is not None:
                    name_element = ET.Element("name")
                    name_element.text = cmd.name
                    cmd_element.append(name_element)

                # Optional list tag
                if cmd.param_list is not None:
                    param_list_element = ET.Element("list")
                    for param in cmd.param_list:
                        param_element = ET.Element("param")
                        param_element.set("name", param.name)
                        param_element.text = param.text
                        param_list_element.append(param_element)
                    cmd_element.append(param_list_element)

                xml_root.append(cmd_element)

                # Optional command parameter
                if cmd.set_command is not None:
                    command_element = ET.Element(cmd.set_command.name)
                    command_element.text = cmd.set_command.text
                    xml_root.append(command_element)

            xml_tree = ET.ElementTree(xml_root)
            # XML declaration only for the first chunk
            xml_tree.write(
                body, encoding="utf-8", xml_declaration=bool(i == 0))

        body_bytes = body.getvalue()

        # Buffered XML not needed anymore: close
        body.close()

        return body_bytes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the interface to Denon AVR receivers.

:copyright: (c) 2016 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import logging
import time

from typing import Dict, List, Optional

import attr
import httpx

from .decorators import run_async_synchronously
from .foundation import DenonAVRFoundation, set_api_host, set_api_timeout
from .const import (
    DENON_ATTR_SETATTR, MAIN_ZONE, VALID_ZONES)
from .exceptions import AvrCommandError

from .audyssey import DenonAVRAudyssey, audyssey_factory
from .input import DenonAVRInput, input_factory
from .soundmode import DenonAVRSoundMode, sound_mode_factory
from .tonecontrol import DenonAVRToneControl, tone_control_factory
from .volume import DenonAVRVolume, volume_factory


_LOGGER = logging.getLogger(__name__)


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVR(DenonAVRFoundation):
    """
    Representing a Denon AVR Device.

    Initialize MainZone of DenonAVR.

    :param host: IP or HOSTNAME.
    :type host: str

    :param name: Device name, if None FriendlyName of device is used.
    :type name: str or None

    :param show_all_inputs: If True deleted input functions are also shown
    :type show_all_inputs: bool

    :param timeout: Timeout when calling device APIs.
    :type timeout: float

    :param add_zones: Additional Zones for which an instance are created
    :type add_zones: dict [str, str] or None
    """

    _host: str = attr.ib(
        converter=str, on_setattr=[*DENON_ATTR_SETATTR, set_api_host])
    _name: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None)
    _show_all_inputs: bool = attr.ib(converter=bool, default=False)
    _add_zones: Optional[Dict[str, str]] = attr.ib(
        validator=attr.validators.optional(attr.validators.deep_mapping(
            attr.validators.in_(VALID_ZONES),
            attr.validators.optional(attr.validators.instance_of(str)),
            attr.validators.instance_of(dict))),
        default=None)
    _timeout: float = attr.ib(
        converter=float,
        on_setattr=[*DENON_ATTR_SETATTR, set_api_timeout],
        default=2.0)
    _zones: Dict[str, DenonAVRFoundation] = attr.ib(
        validator=attr.validators.deep_mapping(
            attr.validators.in_(VALID_ZONES),
            attr.validators.instance_of(DenonAVRFoundation),
            attr.validators.instance_of(dict)),
        default=attr.Factory(dict),
        init=False)
    audyssey: DenonAVRAudyssey = attr.ib(
        validator=attr.validators.instance_of(DenonAVRAudyssey),
        default=attr.Factory(audyssey_factory, takes_self=True),
        init=False)
    input: DenonAVRInput = attr.ib(
        validator=attr.validators.instance_of(DenonAVRInput),
        default=attr.Factory(input_factory, takes_self=True),
        init=False)
    soundmode: DenonAVRSoundMode = attr.ib(
        validator=attr.validators.instance_of(DenonAVRSoundMode),
        default=attr.Factory(sound_mode_factory, takes_self=True),
        init=False)
    tonecontrol: DenonAVRToneControl = attr.ib(
        validator=attr.validators.instance_of(DenonAVRToneControl),
        default=attr.Factory(tone_control_factory, takes_self=True),
        init=False)
    vol: DenonAVRVolume = attr.ib(
        validator=attr.validators.instance_of(DenonAVRVolume),
        default=attr.Factory(volume_factory, takes_self=True),
        init=False)

    def __attrs_post_init__(self) -> None:
        """Initialize special attributes."""
        # Set host and timeout again to start its custom setattr function
        self._host = self._host
        self._timeout = self._timeout

        # Add own instance to zone dictionary
        self._zones[self._device.zone] = self

        # Create instances of additional zones if requested
        if self._device.zone == MAIN_ZONE and self._add_zones is not None:
            self.create_zones(self._add_zones)

    def create_zones(self, add_zones):
        """Create instances of additional zones for the receiver."""
        for zone, zname in add_zones.items():
            # Name either set explicitly or name of Main Zone with suffix
            zonename = None
            if zname is None and self._name is not None:
                zonename = "{} {}".format(self._name, zone)
            zone_device = attr.evolve(self._device, zone=zone)
            zone_inst = DenonAVR(
                host=self._host,
                device=zone_device,
                name=zonename,
                timeout=self._timeout,
                show_all_inputs=self._show_all_inputs)
            self._zones[zone] = zone_inst

    async def async_setup(self) -> None:
        """Ensure that configuration is loaded from receiver asynchronously."""
        # Device setup
        await self._device.async_setup()
        if self._name is None:
            self._name = self._device.friendly_name

        # Setup other functions
        self.input.setup()
        self.soundmode.setup()
        self.tonecontrol.setup()
        self.vol.setup()
        self.audyssey.setup()

        self._is_setup = True

    @run_async_synchronously(async_func=async_setup)
    def setup(self) -> None:
        """Ensure that configuration is loaded from receiver."""

    async def async_update(self):
        """
        Get the latest status information from device asynchronously.

        Method executes the update method for the current receiver type.
        """
        # Ensure that the device is setup
        if self._is_setup is False:
            await self.async_setup()

        # Create a cache id for this global update
        cache_id = time.time()

        # Verify update method
        await self._device.async_verify_avr_2016_update_method(
            cache_id=cache_id)

        # Update device
        await self._device.async_update(global_update=True, cache_id=cache_id)

        # Update other functions
        await self.input.async_update(global_update=True, cache_id=cache_id)
        await self.soundmode.async_update(
            global_update=True, cache_id=cache_id)
        await self.tonecontrol.async_update(
            global_update=True, cache_id=cache_id)
        await self.vol.async_update(global_update=True, cache_id=cache_id)

        # AppCommand0300.xml interface is very slow, thus it is not included
        # into main update
        # await self.audyssey.async_update(
        #     global_update=True, cache_id=cache_id)

    @run_async_synchronously(async_func=async_update)
    def update(self):
        """
        Get the latest status information from device.

        Method executes the update method for the current receiver type.
        """

    async def async_update_tonecontrol(self):
        """Get Tonecontrol settings."""
        await self.tonecontrol.async_update()

    @run_async_synchronously(async_func=async_update_tonecontrol)
    def update_tonecontrol(self):
        """Get Tonecontrol settings."""

    async def async_update_audyssey(self):
        """Get Audyssey settings."""
        await self.audyssey.async_update()

    @run_async_synchronously(async_func=async_update_audyssey)
    def update_audyssey(self):
        """Get Audyssey settings."""

    async def async_get_command(self, request: str) -> str:
        """Send HTTP GET command to Denon AVR receiver asynchronously."""
        return await self._device.api.async_get_command(request)

    @run_async_synchronously(async_func=async_get_command)
    def get_command(self, request: str) -> str:
        """Send HTTP GET command to Denon AVR receiver."""

    @run_async_synchronously(async_func=async_get_command)
    def send_get_command(self, request: str) -> str:
        """Send HTTP GET command to Denon AVR receiver...for compatibility."""

    ##############
    # Properties #
    ##############
    @property
    def zone(self) -> str:
        """Return Zone of this instance."""
        return self._device.zone

    @property
    def zones(self) -> Dict[str, DenonAVRFoundation]:
        """Return all Zone instances of the device."""
        return self._zones

    @property
    def name(self) -> Optional[str]:
        """Return the name of the device as string."""
        return self._name

    @property
    def host(self) -> str:
        """Return the host of the device as string."""
        return self._host

    @property
    def manufacturer(self) -> Optional[str]:
        """Return the manufacturer of the device as string."""
        return self._device.manufacturer

    @property
    def model_name(self) -> Optional[str]:
        """Return the model name of the device as string."""
        return self._device.model_name

    @property
    def serial_number(self) -> Optional[str]:
        """Return the serial number of the device as string."""
        return self._device.serial_number

    @property
    def power(self) -> Optional[str]:
        """
        Return the power state of the device.

        Possible values are: "ON", "STANDBY" and "OFF"
        """
        return self._device.power

    @property
    def state(self) -> Optional[str]:
        """
        Return the state of the device.

        Possible values are: "on", "off", "playing", "paused"
        "playing" and "paused" are only available for input functions
        in PLAYING_SOURCES.
        """
        return self.input.state

    @property
    def muted(self) -> bool:
        """
        Boolean if volume is currently muted.

        Return "True" if muted and "False" if not muted.
        """
        return self.vol.muted

    @property
    def volume(self) -> float:
        """
        Return volume of Denon AVR as float.

        Volume is send in a format like -50.0.
        Minimum is -80.0, maximum at 18.0
        """
        return self.vol.volume

    @property
    def input_func(self) -> Optional[str]:
        """Return the current input source as string."""
        return self.input.input_func

    @property
    def input_func_list(self) -> List[str]:
        """Return a list of available input sources as string."""
        return self.input.input_func_list

    @property
    def support_sound_mode(self) -> Optional[bool]:
        """Return True if sound mode supported."""
        return self.soundmode.support_sound_mode

    @property
    def sound_mode(self) -> Optional[str]:
        """Return the matched current sound mode as a string."""
        return self.soundmode.sound_mode

    @property
    def sound_mode_list(self) -> List[str]:
        """Return a list of available sound modes as string."""
        return self.soundmode.sound_mode_list

    @property
    def sound_mode_map(self) -> Dict[str, str]:  # returns an OrderedDict
        """Return a dict of available sound modes with their mapping values."""
        return self.soundmode.sound_mode_map

    @property
    def sound_mode_map_rev(self) -> Dict[str, str]:
        """Return a dict to map each sound_mode_raw to matching sound_mode."""
        return self.soundmode.sound_mode_map_rev

    @property
    def sound_mode_raw(self) -> Optional[str]:
        """Return the current sound mode as string as received from the AVR."""
        return self.soundmode.sound_mode_raw

    @property
    def image_url(self) -> Optional[str]:
        """Return image URL of current playing media when powered on."""
        return self.input.image_url

    @property
    def title(self) -> Optional[str]:
        """Return title of current playing media as string."""
        return self.input.title

    @property
    def artist(self) -> Optional[str]:
        """Return artist of current playing media as string."""
        return self.input.artist

    @property
    def album(self) -> Optional[str]:
        """Return album name of current playing media as string."""
        return self.input.album

    @property
    def band(self) -> Optional[str]:
        """Return band of current radio station as string."""
        return self.input.band

    @property
    def frequency(self) -> Optional[str]:
        """Return frequency of current radio station as string."""
        return self.input.frequency

    @property
    def station(self) -> Optional[str]:
        """Return current radio station as string."""
        return self.input.station

    @property
    def netaudio_func_list(self) -> List[str]:
        """Return list of network audio devices.

        Those devices should react to play, pause, next and previous
        track commands.
        """
        return self.input.netaudio_func_list

    @property
    def playing_func_list(self) -> List[str]:
        """Return list of playing devices.

        Those devices offer additional information about what they are playing
        (e.g. title, artist, album, band, frequency, station, image_url).
        """
        return self.input.playing_func_list

    @property
    def receiver_port(self) -> int:
        """Return the receiver's port."""
        if self._device.receiver is None:
            return None
        return self._device.receiver.port

    @property
    def receiver_type(self) -> Optional[str]:
        """Return the receiver's type."""
        if self._device.receiver is None:
            return None
        return self._device.receiver.type

    @property
    def show_all_inputs(self) -> Optional[bool]:
        """Indicate if all inputs are shown or just active one."""
        return self._show_all_inputs

    @property
    def bass(self) -> Optional[int]:
        """Return value of bass."""
        return self.tonecontrol.bass

    @property
    def bass_level(self) -> Optional[str]:
        """Return level of bass."""
        return self.tonecontrol.bass_level

    @property
    def treble(self) -> Optional[int]:
        """Return value of treble."""
        return self.tonecontrol.treble

    @property
    def treble_level(self) -> Optional[str]:
        """Return level of treble."""
        return self.tonecontrol.treble_level

    @property
    def dynamic_eq(self) -> Optional[bool]:
        """Return value of Dynamic EQ."""
        return self.audyssey.dynamic_eq

    @property
    def reference_level_offset(self) -> Optional[str]:
        """Return value of Reference Level Offset."""
        return self.audyssey.reference_level_offset

    @property
    def reference_level_offset_setting_list(self) -> List[str]:
        """Return a list of available reference level offset settings."""
        return self.audyssey.reference_level_offset_setting_list

    @property
    def dynamic_volume(self) -> Optional[str]:
        """Return value of Dynamic Volume."""
        return self.audyssey.dynamic_volume

    @property
    def dynamic_volume_setting_list(self) -> List[str]:
        """Return a list of available Dynamic Volume settings."""
        return self.audyssey.dynamic_volume_setting_list

    @property
    def multi_eq(self) -> Optional[str]:
        """Return value of MultiEQ."""
        return self.audyssey.multi_eq

    @property
    def multi_eq_setting_list(self) -> List[str]:
        """Return a list of available MultiEQ settings."""
        return self.audyssey.multi_eq_setting_list

    async def async_dynamic_eq_off(self) -> None:
        """Turn DynamicEQ off."""
        await self.audyssey.async_dynamiceq_off()

    ##########
    # Setter #
    ##########
    def set_async_client(self, async_client: httpx.AsyncClient) -> None:
        """
        Set a custom httpx.AsyncClient for this instance.

        This is a non-blocking method.
        """
        if not issubclass(async_client, httpx.AsyncClient):
            raise AvrCommandError(
                "Provided client is not a subclass of httpx.AsyncClient")
        self._device.api.async_client = async_client

    @run_async_synchronously(async_func=async_dynamic_eq_off)
    def dynamic_eq_off(self) -> None:
        """Turn DynamicEQ off."""

    async def async_dynamic_eq_on(self) -> None:
        """Turn DynamicEQ on."""
        await self.audyssey.async_dynamiceq_on()

    @run_async_synchronously(async_func=async_dynamic_eq_on)
    def dynamic_eq_on(self) -> None:
        """Turn DynamicEQ on."""

    async def async_toggle_dynamic_eq(self) -> None:
        """Toggle DynamicEQ."""
        await self.audyssey.async_toggle_dynamic_eq()

    @run_async_synchronously(async_func=async_toggle_dynamic_eq)
    def toggle_dynamic_eq(self) -> None:
        """Toggle DynamicEQ."""

    async def async_set_input_func(self, input_func: str) -> None:
        """
        Set input_func of device.

        Valid values depend on the device and should be taken from
        "input_func_list".
        """
        await self.input.async_set_input_func(input_func)

    @run_async_synchronously(async_func=async_set_input_func)
    def set_input_func(self, input_func: str) -> None:
        """
        Set input_func of device.

        Valid values depend on the device and should be taken from
        "input_func_list".
        """

    async def async_set_sound_mode(self, sound_mode: str) -> None:
        """
        Set sound_mode of device.

        Valid values depend on the device and should be taken from
        "sound_mode_list".
        """
        await self.soundmode.async_set_sound_mode(sound_mode)

    @run_async_synchronously(async_func=async_set_sound_mode)
    def set_sound_mode(self, sound_mode: str) -> None:
        """
        Set sound_mode of device.

        Valid values depend on the device and should be taken from
        "sound_mode_list".
        """

    async def async_toggle_play_pause(self) -> None:
        """Toggle play pause media player."""
        await self.input.async_toggle_play_pause()

    @run_async_synchronously(async_func=async_toggle_play_pause)
    def toggle_play_pause(self) -> None:
        """Toggle play pause media player."""

    async def async_play(self) -> None:
        """Send play command to receiver command via HTTP post."""
        await self.input.async_play()

    @run_async_synchronously(async_func=async_play)
    def play(self) -> None:
        """Send play command to receiver command via HTTP post."""

    async def async_pause(self) -> None:
        """Send pause command to receiver command via HTTP post."""
        await self.input.async_pause()

    @run_async_synchronously(async_func=async_pause)
    def pause(self) -> None:
        """Send pause command to receiver command via HTTP post."""

    async def async_previous_track(self) -> None:
        """Send previous track command to receiver command via HTTP post."""
        await self.input.async_previous_track()

    @run_async_synchronously(async_func=async_previous_track)
    def previous_track(self) -> None:
        """Send previous track command to receiver command via HTTP post."""

    async def async_next_track(self) -> None:
        """Send next track command to receiver command via HTTP post."""
        await self.input.async_next_track()

    @run_async_synchronously(async_func=async_next_track)
    def next_track(self) -> None:
        """Send next track command to receiver command via HTTP post."""

    async def async_power_on(self) -> None:
        """Turn on receiver via HTTP get command."""
        await self._device.async_power_on()

    @run_async_synchronously(async_func=async_power_on)
    def power_on(self) -> None:
        """Turn on receiver via HTTP get command."""

    async def async_power_off(self) -> None:
        """Turn off receiver via HTTP get command."""
        await self._device.async_power_off()

    @run_async_synchronously(async_func=async_power_off)
    def power_off(self) -> None:
        """Turn off receiver via HTTP get command."""

    async def async_volume_up(self) -> None:
        """Volume up receiver via HTTP get command."""
        await self.vol.async_volume_up()

    @run_async_synchronously(async_func=async_volume_up)
    def volume_up(self) -> None:
        """Volume up receiver via HTTP get command."""

    async def async_volume_down(self) -> None:
        """Volume down receiver via HTTP get command."""
        await self.vol.async_volume_down()

    @run_async_synchronously(async_func=async_volume_down)
    def volume_down(self) -> None:
        """Volume down receiver via HTTP get command."""

    async def async_set_volume(self, volume: float) -> None:
        """
        Set receiver volume via HTTP get command.

        Volume is send in a format like -50.0.
        Minimum is -80.0, maximum at 18.0
        """
        await self.vol.async_set_volume(volume)

    @run_async_synchronously(async_func=async_set_volume)
    def set_volume(self, volume: float) -> None:
        """
        Set receiver volume via HTTP get command.

        Volume is send in a format like -50.0.
        Minimum is -80.0, maximum at 18.0
        """

    async def async_mute(self, mute: bool) -> None:
        """Mute receiver via HTTP get command."""
        await self.vol.async_mute(mute)

    @run_async_synchronously(async_func=async_mute)
    def mute(self, mute: bool) -> None:
        """Mute receiver via HTTP get command."""

    async def async_enable_tone_control(self) -> None:
        """Enable tone control to change settings like bass or treble."""
        await self.tonecontrol.async_enable_tone_control()

    @run_async_synchronously(async_func=async_enable_tone_control)
    def enable_tone_control(self) -> None:
        """Enable tone control to change settings like bass or treble."""

    async def async_disable_tone_control(self) -> None:
        """Disable tone control to change settings like bass or treble."""
        await self.tonecontrol.async_disable_tone_control()

    @run_async_synchronously(async_func=async_disable_tone_control)
    def disable_tone_control(self) -> None:
        """Disable tone control to change settings like bass or treble."""

    async def async_set_bass(self, value: int) -> None:
        """
        Set receiver bass.

        Minimum is 0, maximum at 12

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        await self.tonecontrol.async_set_bass(value)

    @run_async_synchronously(async_func=async_set_bass)
    def set_bass(self, value: int) -> None:
        """
        Set receiver bass.

        Minimum is 0, maximum at 12

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """

    async def async_bass_up(self) -> None:
        """
        Increase level of Bass.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """
        await self.tonecontrol.async_bass_up()

    @run_async_synchronously(async_func=async_bass_up)
    def bass_up(self) -> None:
        """
        Increase level of Bass.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """

    async def async_bass_down(self) -> None:
        """
        Decrease level of Bass.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """
        await self.tonecontrol.async_bass_down()

    @run_async_synchronously(async_func=async_bass_down)
    def bass_down(self) -> None:
        """
        Decrease level of Bass.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """

    async def async_set_treble(self, value: int) -> None:
        """
        Set receiver treble.

        Minimum is 0, maximum at 12

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        await self.tonecontrol.async_set_treble(value)

    @run_async_synchronously(async_func=async_set_treble)
    def set_treble(self, value: int) -> None:
        """
        Set receiver treble.

        Minimum is 0, maximum at 12

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """

    async def async_treble_up(self) -> None:
        """
        Increase level of Treble.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """
        await self.tonecontrol.async_treble_up()

    @run_async_synchronously(async_func=async_treble_up)
    def treble_up(self) -> None:
        """
        Increase level of Treble.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """

    async def async_treble_down(self) -> None:
        """
        Decrease level of Treble.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """
        await self.tonecontrol.async_treble_down()

    @run_async_synchronously(async_func=async_treble_down)
    def treble_down(self) -> None:
        """
        Decrease level of Treble.

        Note:
        Doesn't work, if Dynamic Equalizer is active
        """



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the handler for input functions of Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import logging
import html
import time

from copy import deepcopy
from typing import Dict, Hashable, List, Optional, Tuple

import attr
import httpx

from .appcommand import AppCommands
from .const import (
    ALBUM_COVERS_URL, APPCOMMAND_CMD_TEXT, AVR, AVR_X_2016, AVR_X,
    CHANGE_INPUT_MAPPING, DENON_ATTR_SETATTR, NETAUDIO_SOURCES,
    PLAYING_SOURCES, POWER_ON, SOURCE_MAPPING, STATE_PLAYING, STATE_PAUSED,
    STATIC_ALBUM_URL, ZONE2, ZONE3)
from .exceptions import AvrCommandError, AvrProcessingError, AvrRequestError
from .foundation import DenonAVRFoundation


_LOGGER = logging.getLogger(__name__)


def lower_string(value: Optional[str]) -> Optional[str]:
    """Convert string to lower case."""
    if value is None:
        return value
    return str(value).lower()


def unescape_string(value: Optional[str]) -> Optional[str]:
    """Perform HTML unescape on value."""
    if value is None:
        return value
    return html.unescape(str(value))


def set_input_func(
        instance: DenonAVRFoundation,
        attribute: attr.Attribute,
        value: str) -> str:
    """Set input_func after mapping from input_func_map."""
    # pylint: disable=protected-access
    # No changes for None
    if value is None:
        return value
    # Map from input_func_map
    # AirPlay and Internet Radio are not always listed in available sources
    if value in ["AirPlay", "Internet Radio"]:
        if value not in instance._input_func_map:
            instance._input_func_map[value] = value
            instance._input_func_map_rev[value] = value
    try:
        input_func = instance._input_func_map_rev[value]
    except KeyError:
        _LOGGER.info("No mapping for source %s found", value)
        input_func = value

    return input_func


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRInput(DenonAVRFoundation):
    """This class implements input functions of Denon AVR receiver."""

    _show_all_inputs: bool = attr.ib(converter=bool, default=False)
    _input_func_map: Dict[str, str] = attr.ib(
        validator=attr.validators.deep_mapping(
            attr.validators.instance_of(str),
            attr.validators.instance_of(str)),
        default=attr.Factory(dict))
    _input_func_map_rev: Dict[str, str] = attr.ib(
        validator=attr.validators.deep_mapping(
            attr.validators.instance_of(str),
            attr.validators.instance_of(str)),
        default=attr.Factory(dict))
    _netaudio_func_list: List[str] = attr.ib(
        validator=attr.validators.deep_iterable(
            attr.validators.instance_of(str),
            attr.validators.instance_of(list)),
        default=attr.Factory(list))
    _playing_func_list: List[str] = attr.ib(
        validator=attr.validators.deep_iterable(
            attr.validators.instance_of(str),
            attr.validators.instance_of(list)),
        default=attr.Factory(list))
    _favorite_func_list: List[str] = attr.ib(
        validator=attr.validators.deep_iterable(
            attr.validators.instance_of(str),
            attr.validators.instance_of(list)),
        default=attr.Factory(list))

    _state: Optional[str] = attr.ib(
        converter=attr.converters.optional(lower_string),
        default=None)
    _input_func: Optional[str] = attr.ib(
        converter=attr.converters.optional(str),
        on_setattr=[*DENON_ATTR_SETATTR, set_input_func],
        default=None)

    _artist: Optional[str] = attr.ib(
        converter=attr.converters.optional(unescape_string),
        default=None)
    _album: Optional[str] = attr.ib(
        converter=attr.converters.optional(unescape_string),
        default=None)
    _band: Optional[str] = attr.ib(
        converter=attr.converters.optional(unescape_string),
        default=None)
    _title: Optional[str] = attr.ib(
        converter=attr.converters.optional(unescape_string),
        default=None)
    _frequency: Optional[str] = attr.ib(
        converter=attr.converters.optional(unescape_string),
        default=None)
    _station: Optional[str] = attr.ib(
        converter=attr.converters.optional(unescape_string),
        default=None)
    _image_url: Optional[str] = attr.ib(
        converter=attr.converters.optional(str),
        default=None)
    _image_available: Optional[str] = attr.ib(
        converter=attr.converters.optional(str),
        default=None)

    # Update tags for attributes
    # AppCommand.xml interface
    appcommand_attrs = {
        AppCommands.GetAllZoneSource: None}
    # Status.xml interface
    status_xml_attrs = {
        "_input_func": "./InputFuncSelect/value"}

    def setup(self) -> None:
        """Ensure that the instance is initialized."""
        # Add tags for a potential AppCommand.xml update
        # For update of input function list
        self._device.api.add_appcommand_update_tag(
            AppCommands.GetAllZoneSource)
        self._device.api.add_appcommand_update_tag(
            AppCommands.GetRenameSource)
        self._device.api.add_appcommand_update_tag(
            AppCommands.GetDeletedSource)
        # For update of state
        for tag in self.appcommand_attrs:
            self._device.api.add_appcommand_update_tag(tag)

        self._is_setup = True

    async def async_update(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None) -> None:
        """Update input functions asynchronously."""
        # Ensure instance is setup before updating
        if self._is_setup is False:
            self.setup()

        # Update input functions
        await self.async_update_inputfuncs(
            global_update=global_update, cache_id=cache_id)
        # Update state
        await self.async_update_state(
            global_update=global_update, cache_id=cache_id)
        # Update media state
        await self.async_update_media_state(cache_id=cache_id)

    async def async_get_sources_deviceinfo(self) -> Dict[str, str]:
        """Get sources from Deviceinfo.xml."""
        try:
            # Deviceinfo.xml is static and can be cached for the whole time
            xml = await self._device.api.async_get_xml(
                self._device.urls.deviceinfo, cache_id=0)
        except AvrRequestError as err:
            _LOGGER.debug(
                "Error when getting sources", exc_info=err)
            raise

        receiver_sources = {}
        # Source determination from XML
        favorites = xml.find(".//FavoriteStation")
        if favorites is not None:
            favorite_func_list = []
            for child in favorites:
                if not child.tag.startswith("Favorite"):
                    continue
                func_name = child.tag.upper()
                favorite_func_list.append(func_name)
                receiver_sources[func_name] = child.find("Name").text
            self._favorite_func_list = favorite_func_list
        for xml_zonecapa in xml.findall("DeviceZoneCapabilities"):
            zone_no = "0"
            if self._device.zone == ZONE2:
                zone_no = "1"
            elif self._device.zone == ZONE3:
                zone_no = "2"
            if xml_zonecapa.find("./Zone/No").text == zone_no:
                # Get list of all input sources of receiver
                xml_list = xml_zonecapa.find("./InputSource/List")
                for xml_source in xml_list.findall("Source"):
                    receiver_sources[
                        xml_source.find(
                            "FuncName").text] = xml_source.find(
                                "DefaultName").text

        return receiver_sources

    async def async_get_changed_sources_appcommand(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None
            ) -> Tuple[Dict[str, str], Dict[str, str]]:
        """
        Get renamed and deleted sources lists from receiver .

        Internal method which queries device via HTTP to get names of renamed
        input sources. In this method AppCommand.xml is used.
        """
        # renamed_sources and deleted_sources are dicts with "source" as key
        # and "renamed_source" or deletion flag as value.
        renamed_sources = {}
        deleted_sources = {}

        # Collect tags for AppCommand.xml call
        tags = (AppCommands.GetRenameSource, AppCommands.GetDeletedSource)
        # Execute call
        try:
            if global_update:
                xml = await self._device.api.async_get_global_appcommand(
                    cache_id=cache_id)
            else:
                xml = await self._device.api.async_post_appcommand(
                    self._device.urls.appcommand, tags, cache_id=cache_id)
        except AvrRequestError as err:
            _LOGGER.debug(
                "Error when getting changed sources", exc_info=err)
            raise

        for child in xml.findall(
                "./cmd[@{attribute}='{cmd}']/functionrename/list".format(
                    attribute=APPCOMMAND_CMD_TEXT,
                    cmd=AppCommands.GetRenameSource.cmd_text)):
            try:
                renamed_sources[child.find("name").text.strip()] = (
                    child.find("rename").text.strip())
            except AttributeError:
                continue

        for child in xml.findall(
                "./cmd[@{attribute}='{cmd}']/functiondelete/list".format(
                    attribute=APPCOMMAND_CMD_TEXT,
                    cmd=AppCommands.GetDeletedSource.cmd_text)):
            try:
                deleted_sources[child.find("FuncName").text.strip(
                )] = "DEL" if (
                    child.find("use").text.strip() == "0") else None
            except AttributeError:
                continue

        return (renamed_sources, deleted_sources)

    async def async_get_changed_sources_status_xml(
            self,
            cache_id: Optional[Hashable] = None
            ) -> Tuple[Dict[str, str], Dict[str, str]]:
        """
        Get renamed and deleted sources lists from receiver .

        Internal method which queries device via HTTP to get names of renamed
        input sources.
        """
        # renamed_sources and deleted_sources are dicts with "source" as key
        # and "renamed_source" or deletion flag as value.
        renamed_sources = {}
        deleted_sources = {}
        xml_inputfunclist = []
        xml_renamesource = []
        xml_deletesource = []

        # This XML is needed to get names of eventually renamed sources
        try:
            # AVR-X and AVR-nonX using different XMLs to provide info about
            # deleted sources
            if self._device.receiver == AVR_X:
                xml = await self._device.api.async_get_xml(
                    self._device.urls.status, cache_id=cache_id)
            # URL only available for Main Zone.
            elif self._device.receiver == AVR:
                xml = await self._device.api.async_get_xml(
                   self._device.urls.mainzone, cache_id=cache_id)
            else:
                raise AvrProcessingError(
                    "Method does not work for receiver type {}".format(
                        self._device.receiver.type))
        except AvrRequestError as err:
            _LOGGER.debug(
                "Error when getting changed sources", exc_info=err)
            raise

        # Get the relevant tags from XML structure
        for child in xml:
            # Default names of the sources
            if child.tag == "InputFuncList":
                for value in child:
                    xml_inputfunclist.append(value.text)
            # Renamed sources
            if child.tag == "RenameSource":
                for value in child:
                    # Two different kinds of source structure types exist
                    # 1. <RenameSource><Value>...
                    if value.text is not None:
                        xml_renamesource.append(value.text.strip())
                    # 2. <RenameSource><Value><Value>
                    else:
                        try:
                            xml_renamesource.append(value[0].text.strip())
                        # Exception covers empty tags and appends empty line
                        # in this case, to ensure that sources and
                        # renamed_sources lists have always the same length
                        except IndexError:
                            xml_renamesource.append(None)
            # Deleted sources
            if child.tag == "SourceDelete":
                for value in child:
                    xml_deletesource.append(value.text)

        # If the deleted source list is empty then use all sources.
        if not xml_deletesource:
            xml_deletesource = ['USE'] * len(xml_inputfunclist)

        # Renamed and deleted sources are in the same row as the default ones
        # Only values which are not None are considered. Otherwise translation
        # is not valid and original name is taken
        for i, item in enumerate(xml_inputfunclist):
            try:
                if xml_renamesource[i] is not None:
                    renamed_sources[item] = xml_renamesource[i]
                else:
                    renamed_sources[item] = item
            except IndexError:
                _LOGGER.error(
                    "List of renamed sources incomplete, continuing anyway")
            try:
                deleted_sources[item] = xml_deletesource[i]
            except IndexError:
                _LOGGER.error(
                    "List of deleted sources incomplete, continuing anyway")

        return (renamed_sources, deleted_sources)

    async def async_update_inputfuncs(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None) -> None:
        """Update sources list from receiver."""
        if self._device.receiver in [AVR_X, AVR_X_2016]:
            await self._async_update_inputfuncs_avr_x(
                global_update=global_update, cache_id=cache_id)
        elif self._device.receiver in [AVR]:
            await self._async_update_inputfuncs_avr(cache_id=cache_id)
        else:
            raise AvrProcessingError(
                "Device is not setup correctly, receiver type not set")

    async def _async_update_inputfuncs_avr_x(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None) -> None:
        """
        Update sources list from receiver for AVR-X and AVR-X-2016 devices.

        Internal method which updates sources list of receiver after getting
        sources and potential renaming information from receiver.
        """
        # Get all sources from receiver
        receiver_sources = await self.async_get_sources_deviceinfo()

        if not receiver_sources:
            _LOGGER.debug(
                "Receiver sources list empty. Please check if device is "
                "powered on.")

        # Get renamed and deleted sources
        # From Appcommand.xml if supported
        if self._device.use_avr_2016_update is True:
            renamed_sources, deleted_sources = await (
                self.async_get_changed_sources_appcommand(
                    global_update=global_update, cache_id=cache_id))
        else:
            # Else from status xml
            renamed_sources, deleted_sources = await (
                self.async_get_changed_sources_status_xml(cache_id=cache_id))

        # Remove all deleted sources
        if self._show_all_inputs is False:
            for deleted_source in deleted_sources.items():
                if deleted_source[1] == "DEL":
                    receiver_sources.pop(deleted_source[0], None)

        # Rebuild the sources lists
        input_func_map = {}
        input_func_map_rev = {}
        netaudio_func_list = []
        playing_func_list = []

        for item in receiver_sources.items():
            # Mapping of item[0] because some func names are inconsistant
            # at AVR-X receivers

            m_item_0 = SOURCE_MAPPING.get(item[0], item[0])

            # For renamed sources use those names and save the default name
            # for a later mapping
            if item[0] in renamed_sources:
                input_func_map[renamed_sources[item[0]]] = m_item_0
                input_func_map_rev[
                    m_item_0] = renamed_sources[item[0]]
                # If the source is a netaudio source, save its renamed name
                if item[0] in NETAUDIO_SOURCES:
                    netaudio_func_list.append(
                        renamed_sources[item[0]])
                # If the source is a playing source, save its renamed name
                if item[0] in PLAYING_SOURCES:
                    playing_func_list.append(
                        renamed_sources[item[0]])
            # Otherwise the default names are used
            else:
                input_func_map[item[1]] = m_item_0
                input_func_map_rev[m_item_0] = item[1]
                # If the source is a netaudio source, save its name
                if item[1] in NETAUDIO_SOURCES:
                    netaudio_func_list.append(item[1])
                # If the source is a playing source, save its name
                if item[1] in PLAYING_SOURCES:
                    playing_func_list.append(item[1])

        self._input_func_map = input_func_map
        self._input_func_map_rev = input_func_map_rev
        self._netaudio_func_list = netaudio_func_list
        self._playing_func_list = playing_func_list

    async def _async_update_inputfuncs_avr(
            self,
            cache_id: Optional[Hashable] = None) -> None:
        """
        Update sources list from receiver for AVR devices.

        Internal method which updates sources list of receiver after getting
        sources and potential renaming information from receiver.
        """
        # Get source from status xml
        receiver_sources, deleted_sources = await (
            self.async_get_changed_sources_status_xml(cache_id=cache_id))

        # Remove all deleted sources
        if self._show_all_inputs is False:
            for deleted_source in deleted_sources.items():
                if deleted_source[1] == "DEL":
                    receiver_sources.pop(deleted_source[0], None)

        # Rebuild the sources lists
        input_func_map = {}
        input_func_map_rev = {}
        netaudio_func_list = []
        playing_func_list = []
        for item in receiver_sources.items():
            input_func_map[item[1]] = item[0]
            input_func_map_rev[item[0]] = item[1]
            # If the source is a netaudio source, save its name
            if item[0] in NETAUDIO_SOURCES:
                netaudio_func_list.append(item[1])
            # If the source is a playing source, save its name
            if item[0] in PLAYING_SOURCES:
                playing_func_list.append(item[1])

        self._input_func_map = input_func_map
        self._input_func_map_rev = input_func_map_rev
        self._netaudio_func_list = netaudio_func_list
        self._playing_func_list = playing_func_list

    async def async_update_state(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None):
        """Update state of device."""
        if self._device.use_avr_2016_update is True:
            await self.async_update_attrs_appcommand(
                self.appcommand_attrs, global_update=global_update,
                cache_id=cache_id)
        elif self._device.use_avr_2016_update is False:
            urls = [self._device.urls.status, self._device.urls.mainzone]
            await self.async_update_attrs_status_xml(
                self.status_xml_attrs, urls, cache_id=cache_id)
        else:
            raise AvrProcessingError(
                "Device is not setup correctly, update method not set")

    async def async_update_media_state(
            self,
            cache_id: Optional[Hashable] = None):
        """Update media state of device."""
        # Now playing information is not implemented for 2016+ models, as
        # a HEOS API query needed. So only sync the power state for now.
        if self._device.receiver == AVR_X_2016:
            self._state = self._device.power
        elif (self._device.power == POWER_ON and
                self._input_func in self._playing_func_list):
            await self._async_update_media_data(cache_id)
        else:
            self._state = self._device.power
            self._title = None
            self._artist = None
            self._album = None
            self._band = None
            self._frequency = None
            self._station = None
            self._image_url = None

    async def _async_update_media_data(
            self,
            cache_id: Optional[Hashable] = None):
        """Update media data of device."""
        urls = []
        status_xml_attrs = {}

        # Hash attributes to check for changed track afterwards
        img_change_hash = hash((self._title, self._artist, self._album))

        if self._input_func in self._netaudio_func_list:
            urls = [self._device.urls.netaudiostatus]
            status_xml_attrs = {
                "_title": "./szLine/value[2]",
                "_artist": "./szLine/value[3]",
                "_album": "./szLine/value[5]"}
            self._band = None
            self._frequency = None
            self._station = None
            # Image URL and state are detected after update
        elif self._input_func in ["Tuner", "TUNER"]:
            urls = [self._device.urls.tunerstatus]
            status_xml_attrs = {
                "_band": "./Band/value",
                "_frequency": "./Frequency/value"}
            self._title = None
            self._artist = None
            self._album = None
            self._station = None
            # No special cover, using a static one
            self._image_url = (
                STATIC_ALBUM_URL.format(
                    host=self._device.api.host, port=self._device.api.port))
            # Assume Tuner is always PLAYING
            self._state = STATE_PLAYING
        elif self._input_func in ["HD Radio", "HDRADIO"]:
            urls = [self._device.urls.hdtunerstatus]
            status_xml_attrs = {
                "_artist": "./Artist/value",
                "_album": "./Album/value",
                "_title": "./Title/value",
                "_band": "./Band/value",
                "_frequency": "./Frequency/value",
                "_station": "./StationNameSh /value"}
            # No special cover, using a static one
            self._image_url = (
                STATIC_ALBUM_URL.format(
                    host=self._device.api.host, port=self._device.api.port))
            # Assume Tuner is always PLAYING
            self._state = STATE_PLAYING

        await self.async_update_attrs_status_xml(
            status_xml_attrs, urls, cache_id=cache_id)

        if self._input_func in self._netaudio_func_list:
            if img_change_hash != hash(
                    (self._title, self._artist, self._album)):
                # Refresh cover with a new time stamp for media URL
                # when track is changing
                self._image_url = (ALBUM_COVERS_URL.format(
                    host=self._device.api.host, port=self._device.api.port,
                    time=int(time.time())))
                # On track change assume device is PLAYING
                self._state = STATE_PLAYING

        # Test if image URL is accessable
        if self._image_available is None and self._image_url is not None:
            try:
                async with self._device.api.async_client() as client:
                    res = await client.get(
                        self._image_url, timeout=self._device.api.timeout)
                    res.raise_for_status()
            except httpx.TimeoutException:
                # No result set image URL to None
                self._image_url = None
            except httpx.HTTPStatusError:
                _LOGGER.info('No album art available for your receiver')
                # No image available. Save this status.
                self._image_available = False
                #  Set image URL to None.
                self._image_url = None
            else:
                self._image_available = True
        # Already tested that image URL is not accessible
        elif self._image_available is False:
            self._image_url = None

    ##############
    # Properties #
    ##############
    @property
    def state(self) -> Optional[str]:
        """
        Return the state of the device.

        Possible values are: "on", "off", "playing", "paused"
        "playing" and "paused" are only available for input functions
        in PLAYING_SOURCES.
        """
        return self._state

    @property
    def input_func(self) -> Optional[str]:
        """Return the current input source as string."""
        return self._input_func

    @property
    def input_func_list(self) -> List[str]:
        """Return a list of available input sources as string."""
        return sorted(self._input_func_map.keys())

    @property
    def image_url(self) -> Optional[str]:
        """Return image URL of current playing media when powered on."""
        if self._device.power == POWER_ON:
            return self._image_url
        return None

    @property
    def title(self) -> Optional[str]:
        """Return title of current playing media as string."""
        return self._title

    @property
    def artist(self) -> Optional[str]:
        """Return artist of current playing media as string."""
        return self._artist

    @property
    def album(self) -> Optional[str]:
        """Return album name of current playing media as string."""
        return self._album

    @property
    def band(self) -> Optional[str]:
        """Return band of current radio station as string."""
        return self._band

    @property
    def frequency(self) -> Optional[str]:
        """Return frequency of current radio station as string."""
        return self._frequency

    @property
    def station(self) -> Optional[str]:
        """Return current radio station as string."""
        return self._station

    @property
    def netaudio_func_list(self) -> List[str]:
        """Return list of network audio devices.

        Those devices should react to play, pause, next and previous
        track commands.
        """
        return deepcopy(self._netaudio_func_list)

    @property
    def playing_func_list(self) -> List[str]:
        """Return list of playing devices.

        Those devices offer additional information about what they are playing
        (e.g. title, artist, album, band, frequency, station, image_url).
        """
        return deepcopy(self._playing_func_list)

    ##########
    # Setter #
    ##########

    async def async_set_input_func(self, input_func: str) -> None:
        """
        Set input_func of device.

        Valid values depend on the device and should be taken from
        "input_func_list".
        """
        # For selection of sources other names then at receiving sources
        # have to be used
        # AVR-X receiver needs source mapping to set input_func
        linp = ""
        direct_mapping = True
        if self._device.receiver in [AVR_X, AVR_X_2016]:
            try:
                linp = CHANGE_INPUT_MAPPING[self._input_func_map[input_func]]
            except KeyError:
                direct_mapping = True
            else:
                direct_mapping = False
        # AVR-nonX receiver and if no mapping was found get parameter for
        # setting input_func directly
        if direct_mapping is True:
            try:
                linp = self._input_func_map[input_func]
            except KeyError as err:
                raise AvrCommandError(
                    "No mapping for input source {}".format(
                        input_func)) from err
        # Create command URL and send command via HTTP GET
        if linp in self._favorite_func_list:
            command_url = self._device.urls.command_fav_src + linp
        else:
            command_url = self._device.urls.command_sel_src + linp

        await self._device.api.async_get_command(command_url)

    async def async_toggle_play_pause(self) -> None:
        """Toggle play pause media player."""
        # Use Play/Pause button only for sources which support NETAUDIO
        if (self._state == STATE_PLAYING and
                self._input_func in self._netaudio_func_list):
            await self.async_pause()
        elif self._input_func in self._netaudio_func_list:
            await self.async_play()

    async def async_play(self) -> None:
        """Send play command to receiver command via HTTP post."""
        # Use pause command only for sources which support NETAUDIO
        if self._input_func in self._netaudio_func_list:
            # In fact play command is a play/pause toggle. Thus checking state
            if self._state == STATE_PLAYING:
                _LOGGER.info("Already playing, play command not sent")
                return

            await self._device.api.async_get_command(
                self._device.urls.command_play)
            self._state = STATE_PLAYING

    async def async_pause(self) -> None:
        """Send pause command to receiver command via HTTP post."""
        # Use pause command only for sources which support NETAUDIO
        if self._input_func in self._netaudio_func_list:
            await self._device.api.async_get_command(
                self._device.urls.command_pause)
            self._state = STATE_PAUSED

    async def async_previous_track(self) -> None:
        """Send previous track command to receiver command via HTTP post."""
        # Use previous track button only for sources which support NETAUDIO
        if self._input_func in self._netaudio_func_list:
            body = {"cmd0": "PutNetAudioCommand/CurUp",
                    "cmd1": "aspMainZone_WebUpdateStatus/",
                    "ZoneName": "MAIN ZONE"}
            await self._device.api.async_post(
                self._device.urls.command_netaudio_post, data=body)

    async def async_next_track(self) -> None:
        """Send next track command to receiver command via HTTP post."""
        # Use next track button only for sources which support NETAUDIO
        if self._input_func in self._netaudio_func_list:
            body = {"cmd0": "PutNetAudioCommand/CurDown",
                    "cmd1": "aspMainZone_WebUpdateStatus/",
                    "ZoneName": "MAIN ZONE"}
            await self._device.api.async_post(
                self._device.urls.command_netaudio_post, data=body)


def input_factory(instance: DenonAVRFoundation) -> DenonAVRInput:
    """Create DenonAVRInput at receiver instances."""
    # pylint: disable=protected-access
    new = DenonAVRInput(
        device=instance._device,
        show_all_inputs=instance._show_all_inputs)
    return new



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the handler for state of Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import logging
import time

from typing import Hashable, Optional

import attr

from .appcommand import AppCommandCmdParam, AppCommands
from .const import DENON_ATTR_SETATTR
from .exceptions import AvrProcessingError, AvrCommandError
from .foundation import DenonAVRFoundation, convert_string_int_bool


_LOGGER = logging.getLogger(__name__)


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRToneControl(DenonAVRFoundation):
    """This class implements tone control functions of Denon AVR receiver."""

    _tone_control_status: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool),
        default=None)
    _tone_control_adjust: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool),
        default=None)
    _bass: Optional[int] = attr.ib(
        converter=attr.converters.optional(int),
        default=None)
    _bass_level: Optional[str] = attr.ib(
        converter=attr.converters.optional(str),
        default=None)
    _treble: Optional[int] = attr.ib(
        converter=attr.converters.optional(int),
        default=None)
    _treble_level: Optional[str] = attr.ib(
        converter=attr.converters.optional(str),
        default=None)

    # Update tags for attributes
    # AppCommand.xml interface
    appcommand_attrs = {
        AppCommands.GetToneControl: None}

    def setup(self) -> None:
        """Ensure that the instance is initialized."""
        # Add tags for a potential AppCommand.xml update
        for tag in self.appcommand_attrs:
            self._device.api.add_appcommand_update_tag(tag)

        self._is_setup = True

    async def async_update(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None) -> None:
        """Update volume asynchronously."""
        # Ensure instance is setup before updating
        if self._is_setup is False:
            self.setup()

        # Update state
        await self.async_update_tone_control(
            global_update=global_update, cache_id=cache_id)

    async def async_update_tone_control(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None):
        """Update tone control status of device."""
        if self._device.use_avr_2016_update is True:
            await self.async_update_attrs_appcommand(
                self.appcommand_attrs, global_update=global_update,
                cache_id=cache_id)
        elif self._device.use_avr_2016_update is False:
            # Not available
            pass
        else:
            raise AvrProcessingError(
                "Device is not setup correctly, update method not set")

    async def async_set_tone_control_command(
            self, parameter_type: str, value: int) -> None:
        """Post request for tone control commands."""
        cmd = (attr.evolve(
            AppCommands.SetToneControl,
            set_command=AppCommandCmdParam(name=parameter_type, text=value)),)
        await self._device.api.async_post_appcommand(
            self._device.urls.appcommand, cmd, cache_id=time.time())

    ##############
    # Properties #
    ##############
    @property
    def bass(self) -> Optional[int]:
        """Return value of bass."""
        return self._bass

    @property
    def bass_level(self) -> Optional[str]:
        """Return level of bass."""
        return self._bass_level

    @property
    def treble(self) -> Optional[int]:
        """Return value of treble."""
        return self._treble

    @property
    def treble_level(self) -> Optional[str]:
        """Return level of treble."""
        return self._treble_level

    ##########
    # Setter #
    ##########
    async def async_enable_tone_control(self) -> None:
        """Enable tone control to change settings like bass or treble."""
        if self._tone_control_status is False:
            raise AvrCommandError(
                "Cannot enable tone control, Dynamic EQ must be deactivated")

        await self.async_set_tone_control_command("adjust", 1)

    async def async_disable_tone_control(self) -> None:
        """Disable tone control to change settings like bass or treble."""
        if self._tone_control_status is False:
            raise AvrCommandError(
                "Cannot disable tone control, Dynamic EQ must be deactivated")

        await self.async_set_tone_control_command("adjust", 0)

    async def async_set_bass(self, value: int) -> None:
        """
        Set receiver bass.

        Minimum is 0, maximum at 12

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if value < 0 or value > 12:
            raise AvrCommandError("Invalid value for bass")
        await self.async_enable_tone_control()
        await self.async_set_tone_control_command("bassvalue", value)

    async def async_bass_up(self) -> None:
        """
        Increase level of Bass.

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if self.bass == 12:
            return
        await self.async_enable_tone_control()
        await self.async_set_tone_control_command("bassvalue", self.bass + 1)
        await self.async_update()

    async def async_bass_down(self) -> None:
        """
        Decrease level of Bass.

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if self.bass == 0:
            return
        await self.async_enable_tone_control()
        await self.async_set_tone_control_command("bassvalue", self.bass - 1)
        await self.async_update()

    async def async_set_treble(self, value: int) -> None:
        """
        Set receiver treble.

        Minimum is 0, maximum at 12

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if value < 0 or value > 12:
            raise AvrCommandError("Invalid value for treble")
        await self.async_enable_tone_control()
        await self.async_set_tone_control_command("treblevalue", value)

    async def async_treble_up(self) -> None:
        """
        Increase level of Treble.

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if self.treble == 12:
            return
        await self.async_enable_tone_control()
        await self.async_set_tone_control_command(
            "treblevalue", self.treble + 1
            )
        await self.async_update()

    async def async_treble_down(self) -> None:
        """
        Decrease level of Treble.

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if self.treble == 0:
            return
        await self.async_enable_tone_control()
        await self.async_set_tone_control_command(
            "treblevalue", self.treble - 1
            )
        await self.async_update()


def tone_control_factory(instance: DenonAVRFoundation) -> DenonAVRToneControl:
    """Create DenonAVRToneControl at receiver instances."""
    # pylint: disable=protected-access
    new = DenonAVRToneControl(device=instance._device)
    return new



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for ssdp."""

import pytest
from denonavr.ssdp import evaluate_scpd_xml


def get_sample_content(filename):
    """Return sample content form file."""
    with open(
            "tests/xml/{filename}".format(
                filename=filename), encoding="utf-8") as file:
        return file.read()


@pytest.mark.parametrize(
    "model,expected_device",
    [
        (
            "AVR-X1600H",
            {
                "friendlyName": "Denon AVR-X1600H",
                "host": "10.0.0.0",
                "manufacturer": "Denon",
                "modelName": "Denon AVR-X1600H",
                "serialNumber": "XXX",
            },
        )
    ],
)
def test_evaluate(model, expected_device):
    """Test that the discovered device looks like expected."""
    url = "https://10.0.0.0/denon"
    body = get_sample_content("{model}_upnp.xml".format(model=model))
    device = evaluate_scpd_xml(url, body)
    assert device == expected_device



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the REST API to Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import asyncio
import inspect
import logging
import time
import xml.etree.ElementTree as ET

from functools import wraps
from typing import Callable, Coroutine

import httpx

from defusedxml import DefusedXmlException
from defusedxml.ElementTree import ParseError

from .exceptions import (
    AvrRequestError,
    AvrForbiddenError,
    AvrTimoutError,
    AvrInvalidResponseError)

_LOGGER = logging.getLogger(__name__)


def async_handle_receiver_exceptions(func: Coroutine) -> Coroutine:
    """
    Handle exceptions raised when calling an Denon AVR endpoint asynchronously.

    The decorated function must either have a string variable as second
    argument or as "request" keyword argument.
    """
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except httpx.HTTPStatusError as err:
            _LOGGER.debug(
                "HTTP status error on request %s", err.request, exc_info=True)
            # Separate handling of 403 errors
            if err.response.status_code == 403:
                raise AvrForbiddenError(
                    "HTTPStatusError: {}".format(err), err.request) from err
            raise AvrRequestError(
                "HTTPStatusError: {}".format(err), err.request) from err
        except httpx.TimeoutException as err:
            _LOGGER.debug(
                "HTTP timeout exception on request %s", err.request,
                exc_info=True)
            raise AvrTimoutError(
                "TimeoutException: {}".format(err), err.request) from err
        except (ET.ParseError, DefusedXmlException, ParseError) as err:
            _LOGGER.debug(
                "Defusedxml parse error on request %s", (args, kwargs),
                exc_info=True)
            raise AvrInvalidResponseError(
                "XMLParseError: {}".format(err), (args, kwargs)) from err

    return wrapper


def set_cache_id(func: Callable) -> Callable:
    """
    Decorate a function to add cache_id keyword argument if it is not present.

    The function must be called with a fix cache_id keyword argument to be able
    to get cached data. This prevents accidential caching of a function result.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs.get("cache_id") is None:
            kwargs["cache_id"] = time.time()
        return func(*args, **kwargs)

    return wrapper


def run_async_synchronously(async_func: Coroutine) -> Callable:
    """
    Decorate to run the configured asynchronous function synchronously instead.

    If available the corresponding function with async_ prefix is called in an
    own event loop. This is not efficient but it ensures backwards
    compatibility of this library.
    """
    def decorator(func: Callable):
        # Check if function is a coroutine
        if not inspect.iscoroutinefunction(async_func):
            raise AttributeError(
                "Function {} is not a coroutine function".format(async_func))
        # Check if the signature of both functions is equal
        if inspect.signature(func) != inspect.signature(async_func):
            raise AttributeError(
                "Functions {} and {} have different signatures".format(
                    func, async_func))

        @wraps(func)
        def wrapper(*args, **kwargs):
            # Run async function in own event loop
            loop = asyncio.new_event_loop()

            try:
                return loop.run_until_complete(async_func(*args, **kwargs))
            finally:
                loop.close()

        return wrapper

    return decorator



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements a discovery function for Denon AVR receivers.

:copyright: (c) 2016 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import logging
import asyncio
import socket
import re
import xml.etree.ElementTree as ET

from typing import Dict, List, Optional, Tuple, Set
from urllib.parse import urlparse

import httpx
import netifaces

from defusedxml.ElementTree import fromstring

_LOGGER = logging.getLogger(__name__)

AsyncClient = httpx.AsyncClient

SSDP_ADDR = "239.255.255.250"
SSDP_PORT = 1900
SSDP_MX = 2
SSDP_TARGET = (SSDP_ADDR, SSDP_PORT)
SSDP_ST_1 = "ssdp:all"
SSDP_ST_2 = "upnp:rootdevice"
SSDP_ST_3 = "urn:schemas-upnp-org:device:MediaRenderer:1"

SSDP_ST_LIST = (SSDP_ST_1, SSDP_ST_2, SSDP_ST_3)

SSDP_LOCATION_PATTERN = re.compile(r'(?<=LOCATION:\s).+?(?=\r)')

SCPD_XMLNS = "{urn:schemas-upnp-org:device-1-0}"
SCPD_DEVICE = "{xmlns}device".format(xmlns=SCPD_XMLNS)
SCPD_DEVICELIST = "{xmlns}deviceList".format(xmlns=SCPD_XMLNS)
SCPD_DEVICETYPE = "{xmlns}deviceType".format(xmlns=SCPD_XMLNS)
SCPD_MANUFACTURER = "{xmlns}manufacturer".format(xmlns=SCPD_XMLNS)
SCPD_MODELNAME = "{xmlns}modelName".format(xmlns=SCPD_XMLNS)
SCPD_SERIALNUMBER = "{xmlns}serialNumber".format(xmlns=SCPD_XMLNS)
SCPD_FRIENDLYNAME = "{xmlns}friendlyName".format(xmlns=SCPD_XMLNS)
SCPD_PRESENTATIONURL = "{xmlns}presentationURL".format(xmlns=SCPD_XMLNS)

SUPPORTED_DEVICETYPES = [
    "urn:schemas-upnp-org:device:MediaRenderer:1",
    "urn:schemas-upnp-org:device:MediaServer:1",
    ]

SUPPORTED_MANUFACTURERS = ["Denon", "DENON", "DENON PROFESSIONAL", "Marantz"]


def ssdp_request(ssdp_st: str, ssdp_mx: float = SSDP_MX) -> bytes:
    """Return request bytes for given st and mx."""
    return "\r\n".join([
        'M-SEARCH * HTTP/1.1',
        'ST: {}'.format(ssdp_st),
        'MX: {:d}'.format(ssdp_mx),
        'MAN: "ssdp:discover"',
        'HOST: {}:{}'.format(*SSDP_TARGET),
        '', '']).encode('utf-8')


def get_local_ips() -> List[str]:
    """Get IPs of local network adapters."""
    ips = []
    # pylint: disable=c-extension-no-member
    for interface in netifaces.interfaces():
        addresses = netifaces.ifaddresses(interface)
        for address in addresses.get(netifaces.AF_INET, []):
            ips.append(address["addr"])
    return ips


async def async_identify_denonavr_receivers() -> List[Dict]:
    """
    Identify DenonAVR using SSDP and SCPD queries.

    Returns a list of dictionaries which includes all discovered Denon AVR
    devices with keys "host", "modelName", "friendlyName", "presentationURL".
    """
    # Sending SSDP broadcast message to get resource urls from devices
    urls = await async_send_ssdp_broadcast()

    # Check which responding device is a DenonAVR device and prepare output
    receivers = []

    for url in urls:
        try:
            async with AsyncClient() as client:
                res = await client.get(url, timeout=5.0)
                res.raise_for_status()
        except httpx.HTTPError:
            continue
        else:
            receiver = evaluate_scpd_xml(url, res.text)
            if receiver is not None:
                receivers.append(receiver)

    return receivers


async def async_send_ssdp_broadcast() -> Set[str]:
    """
    Send SSDP broadcast messages to discover UPnP devices.

    Returns a set of SCPD XML resource urls for all discovered devices.
    """
    # Send up to three different broadcast messages
    ips = get_local_ips()
    # Prepare output of responding devices
    urls = set()

    tasks = []
    for ip_addr in ips:
        tasks.append(async_send_ssdp_broadcast_ip(ip_addr))

    results = await asyncio.gather(*tasks)

    for result in results:
        _LOGGER.debug("SSDP broadcast result received: %s", result)
        urls = urls.union(result)

    _LOGGER.debug("Following devices found: %s", urls)
    return urls


async def async_send_ssdp_broadcast_ip(ip_addr: str) -> Set[str]:
    """Send SSDP broadcast messages to a single IP."""
    # Ignore 169.254.0.0/16 adresses
    if re.search("169.254.*.*", ip_addr):
        return set()

    # Prepare socket
    sock = socket.socket(
        socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
    sock.bind((ip_addr, 0))

    # Get asyncio loop
    loop = asyncio.get_event_loop()
    transport, protocol = await loop.create_datagram_endpoint(
        DenonAVRSSDP, sock=sock)

    # Wait for the timeout period
    await asyncio.sleep(SSDP_MX)

    # Close the connection
    transport.close()

    _LOGGER.debug(
        "Got %s results after SSDP queries using ip %s",
        len(protocol.urls), ip_addr)

    return protocol.urls


def evaluate_scpd_xml(url: str, body: str) -> Optional[Dict]:
    """
    Evaluate SCPD XML.

    Returns dictionary with keys "host", "modelName", "friendlyName" and
    "presentationURL" if a Denon AVR device was found and "None" if not.
    """
    try:
        root = fromstring(body)
        # Look for manufacturer "Denon" in response.
        # Using "try" in case tags are not available in XML
        device = {}
        device_xml = None
        device["manufacturer"] = (
            root.find(SCPD_DEVICE).find(SCPD_MANUFACTURER).text)

        _LOGGER.debug(
            "Device %s has manufacturer %s", url, device["manufacturer"])

        if not device["manufacturer"] in SUPPORTED_MANUFACTURERS:
            return None

        if (root.find(SCPD_DEVICE).find(SCPD_DEVICETYPE).text
                in SUPPORTED_DEVICETYPES):
            device_xml = root.find(SCPD_DEVICE)
        elif root.find(SCPD_DEVICE).find(SCPD_DEVICELIST) is not None:
            for dev in root.find(SCPD_DEVICE).find(SCPD_DEVICELIST):
                if (dev.find(SCPD_DEVICETYPE).text in SUPPORTED_DEVICETYPES
                        and dev.find(SCPD_SERIALNUMBER) is not None):
                    device_xml = dev
                    break

        if device_xml is None:
            return None

        if device_xml.find(SCPD_PRESENTATIONURL) is not None:
            device["host"] = urlparse(
                device_xml.find(
                    SCPD_PRESENTATIONURL).text).hostname
            device["presentationURL"] = (
                device_xml.find(SCPD_PRESENTATIONURL).text)
        else:
            device["host"] = urlparse(url).hostname

        device["modelName"] = (
            device_xml.find(SCPD_MODELNAME).text)
        device["serialNumber"] = (
            device_xml.find(SCPD_SERIALNUMBER).text)
        device["friendlyName"] = (
            device_xml.find(SCPD_FRIENDLYNAME).text)
        return device
    except (AttributeError, ValueError, ET.ParseError) as err:
        _LOGGER.error(
            "Error occurred during evaluation of SCPD XML: %s", err)
        return None


class DenonAVRSSDP(asyncio.DatagramProtocol):
    """Implements datagram protocol for SSDP discovery of Denon AVR devices."""

    def __init__(self) -> None:
        """Create instance."""
        self.urls = set()

    def connection_made(
            self, transport: asyncio.DatagramTransport) -> None:
        """Send SSDP request when connection was made."""
        # Prepare SSDP and send broadcast message
        for ssdp_st in SSDP_ST_LIST:
            request = ssdp_request(ssdp_st)
            transport.sendto(request, SSDP_TARGET)
            _LOGGER.debug("SSDP request sent %s", request)

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        """Receive responses to SSDP call."""
        # Some string operations to get the receivers URL
        # which could be found between LOCATION and end of line of the response
        _LOGGER.debug("Response to SSDP call received: %s", data)
        data_text = data.decode("utf-8")
        match = SSDP_LOCATION_PATTERN.search(data_text)
        if match:
            self.urls.add(match.group(0))



#!/usr/bin/env python3
# encoding: utf-8
"""Automation Library for Denon AVR receivers."""
from setuptools import find_packages, setup

setup(name='denonavr',
      version='0.10.2',
      description='Automation Library for Denon AVR receivers',
      long_description='Automation Library for Denon AVR receivers',
      url='https://github.com/scarface-4711/denonavr',
      author='Oliver Goetz',
      author_email='scarface@mywoh.de',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'async_lru>=1.0.2',
          'attrs>=20.3.0',
          'defusedxml>=0.6.0',
          'httpx>=0.16.1',
          'netifaces>=0.10.9'],
      tests_require=['tox'],
      platforms=['any'],
      zip_safe=False,
      python_requires=">=3.6",
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Topic :: Software Development :: Libraries",
          "Topic :: Home Automation",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9"
          ])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the handler for volume of Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import logging

from typing import Hashable, Optional, Union

import attr

from .appcommand import AppCommands
from .const import DENON_ATTR_SETATTR, STATE_ON
from .exceptions import AvrCommandError, AvrProcessingError
from .foundation import DenonAVRFoundation


_LOGGER = logging.getLogger(__name__)


def convert_muted(value: str) -> bool:
    """Convert muted to bool."""
    return bool(value == STATE_ON)


def convert_volume(value: Union[float, str]) -> float:
    """Convert volume to float."""
    if value == "--":
        return -80.0
    return float(value)


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRVolume(DenonAVRFoundation):
    """This class implements volume functions of Denon AVR receiver."""

    _volume: Optional[float] = attr.ib(
        converter=attr.converters.optional(convert_volume),
        default=None)
    _muted: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_muted),
        default=None)

    # Update tags for attributes
    # AppCommand.xml interface
    appcommand_attrs = {
        AppCommands.GetAllZoneVolume: None,
        AppCommands.GetAllZoneMuteStatus: None}
    # Status.xml interface
    status_xml_attrs = {
        "_volume": "./MasterVolume/value",
        "_muted": "./Mute/value"}

    def setup(self) -> None:
        """Ensure that the instance is initialized."""
        # Add tags for a potential AppCommand.xml update
        for tag in self.appcommand_attrs:
            self._device.api.add_appcommand_update_tag(tag)

        self._is_setup = True

    async def async_update(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None) -> None:
        """Update volume asynchronously."""
        # Ensure instance is setup before updating
        if self._is_setup is False:
            self.setup()

        # Update state
        await self.async_update_volume(
            global_update=global_update, cache_id=cache_id)

    async def async_update_volume(
            self,
            global_update: bool = False,
            cache_id: Optional[Hashable] = None):
        """Update volume status of device."""
        if self._device.use_avr_2016_update is True:
            await self.async_update_attrs_appcommand(
                self.appcommand_attrs, global_update=global_update,
                cache_id=cache_id)
        elif self._device.use_avr_2016_update is False:
            urls = [self._device.urls.status, self._device.urls.mainzone]
            await self.async_update_attrs_status_xml(
                self.status_xml_attrs, urls, cache_id=cache_id)
        else:
            raise AvrProcessingError(
                "Device is not setup correctly, update method not set")

    ##############
    # Properties #
    ##############
    @property
    def muted(self) -> Optional[bool]:
        """
        Boolean if volume is currently muted.

        Return "True" if muted and "False" if not muted.
        """
        return self._muted

    @property
    def volume(self) -> Optional[float]:
        """
        Return volume of Denon AVR as float.

        Volume is send in a format like -50.0.
        Minimum is -80.0, maximum at 18.0
        """
        return self._volume

    ##########
    # Setter #
    ##########
    async def async_volume_up(self) -> None:
        """Volume up receiver via HTTP get command."""
        await self._device.api.async_get_command(
            self._device.urls.command_volume_up)

    async def async_volume_down(self) -> None:
        """Volume down receiver via HTTP get command."""
        await self._device.api.async_get_command(
            self._device.urls.command_volume_down)

    async def async_set_volume(self, volume: float) -> None:
        """
        Set receiver volume via HTTP get command.

        Volume is send in a format like -50.0.
        Minimum is -80.0, maximum at 18.0
        """
        if volume < -80 or volume > 18:
            raise AvrCommandError("Invalid volume: {}".format(volume))

        await self._device.api.async_get_command(
            self._device.urls.command_set_volume.format(volume=volume))

    async def async_mute(self, mute: bool) -> None:
        """Mute receiver via HTTP get command."""
        if mute:
            await self._device.api.async_get_command(
                self._device.urls.command_mute_on)
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_mute_off)


def volume_factory(instance: DenonAVRFoundation) -> DenonAVRVolume:
    """Create DenonAVRVolume at receiver instances."""
    # pylint: disable=protected-access
    new = DenonAVRVolume(device=instance._device)
    return new
